# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
The module for working with displaying progress bars for Modin execution engines.

Modin Automatic Query Progress (AQP).
"""

import inspect
import os
import threading
import time
import warnings

from modin.config import Engine, ProgressBar

progress_bars = {}
bar_lock = threading.Lock()


def call_progress_bar(result_parts, line_no):
    """
    Attach a progress bar to given `result_parts`.

    The progress bar is expected to be shown in a Jupyter Notebook cell.

    Parameters
    ----------
    result_parts : list of list of object refs (futures)
        Objects which are being computed for which progress is requested.
    line_no : int
        Line number in the call stack which we're displaying progress for.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            from tqdm.autonotebook import tqdm as tqdm_notebook
        except ImportError:
            raise ImportError("Please pip install tqdm to use the progress bar")
        from IPython import get_ipython

    try:
        cell_no = get_ipython().execution_count
    # This happens if we are not in ipython or jupyter.
    # No progress bar is supported in that case.
    except AttributeError:
        return
    pbar_id = f"{cell_no}-{line_no}"
    futures = [
        block
        for row in result_parts
        for partition in row
        for block in partition.list_of_blocks
    ]
    bar_format = (
        "{l_bar}{bar}{r_bar}"
        if "DEBUG_PROGRESS_BAR" in os.environ
        and os.environ["DEBUG_PROGRESS_BAR"] == "True"
        else "{desc}: {percentage:3.0f}%{bar} Elapsed time: {elapsed}, estimated remaining time: {remaining}"
    )
    bar_lock.acquire()
    if pbar_id in progress_bars:
        if hasattr(progress_bars[pbar_id], "container"):
            if hasattr(progress_bars[pbar_id].container.children[0], "max"):
                index = 0
            else:
                index = 1
            progress_bars[pbar_id].container.children[index].max = progress_bars[
                pbar_id
            ].container.children[index].max + len(futures)
        progress_bars[pbar_id].total = progress_bars[pbar_id].total + len(futures)
        progress_bars[pbar_id].refresh()
    else:
        progress_bars[pbar_id] = tqdm_notebook(
            total=len(futures),
            desc="Estimated completion of line " + str(line_no),
            bar_format=bar_format,
        )
    bar_lock.release()

    threading.Thread(target=_show_time_updates, args=(progress_bars[pbar_id],)).start()

    modin_engine = Engine.get()
    engine_wrapper = None
    if modin_engine == "Ray":
        from modin.core.execution.ray.common.engine_wrapper import RayWrapper

        engine_wrapper = RayWrapper
    elif modin_engine == "Unidist":
        from modin.core.execution.unidist.common.engine_wrapper import UnidistWrapper

        engine_wrapper = UnidistWrapper
    else:
        raise NotImplementedError(
            f"ProgressBar feature is not supported for {modin_engine} engine."
        )

    for i in range(1, len(futures) + 1):
        engine_wrapper.wait(futures, num_returns=i)
        progress_bars[pbar_id].update(1)
        progress_bars[pbar_id].refresh()
    if progress_bars[pbar_id].n == progress_bars[pbar_id].total:
        progress_bars[pbar_id].close()


def display_time_updates(bar):
    """
    Start displaying the progress `bar` in a notebook.

    Parameters
    ----------
    bar : tqdm.tqdm
        The progress bar wrapper to display in a notebook cell.
    """
    threading.Thread(target=_show_time_updates, args=(bar,)).start()


def _show_time_updates(p_bar):
    """
    Refresh displayed progress bar `p_bar` periodically until it is complete.

    Parameters
    ----------
    p_bar : tqdm.tqdm
        The progress bar wrapper being displayed to refresh.
    """
    while p_bar.total > p_bar.n:
        time.sleep(1)
        if p_bar.total > p_bar.n:
            p_bar.refresh()


def progress_bar_wrapper(f):
    """
    Wrap computation function inside a progress bar.

    Spawns another thread which displays a progress bar showing
    estimated completion time.

    Parameters
    ----------
    f : callable
        The name of the function to be wrapped.

    Returns
    -------
    callable
        Decorated version of `f` which reports progress.
    """
    from functools import wraps

    @wraps(f)
    def magic(*args, **kwargs):
        result_parts = f(*args, **kwargs)
        if ProgressBar.get():
            current_frame = inspect.currentframe()
            function_name = None
            while function_name != "<module>":
                (
                    filename,
                    line_number,
                    function_name,
                    lines,
                    index,
                ) = inspect.getframeinfo(current_frame)
                current_frame = current_frame.f_back
            t = threading.Thread(
                target=call_progress_bar,
                args=(result_parts, line_number),
            )
            t.start()
            # We need to know whether or not we are in a jupyter notebook
            from IPython import get_ipython

            try:
                ipy_str = str(type(get_ipython()))
                if "zmqshell" not in ipy_str:
                    t.join()
            except Exception:
                pass
        return result_parts

    return magic



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Implement experimental I/O public API."""

from __future__ import annotations

import inspect
import pathlib
import pickle
from typing import IO, AnyStr, Callable, Iterator, Literal, Optional, Union

import pandas
import pandas._libs.lib as lib
from pandas._typing import CompressionOptions, DtypeArg, DtypeBackend, StorageOptions

from modin.core.storage_formats import BaseQueryCompiler
from modin.utils import expanduser_path_arg

from . import DataFrame


def read_sql(
    sql,
    con,
    index_col=None,
    coerce_float=True,
    params=None,
    parse_dates=None,
    columns=None,
    chunksize=None,
    dtype_backend=lib.no_default,
    dtype=None,
    partition_column: Optional[str] = None,
    lower_bound: Optional[int] = None,
    upper_bound: Optional[int] = None,
    max_sessions: Optional[int] = None,
) -> Union[DataFrame, Iterator[DataFrame]]:
    """
    General documentation is available in `modin.pandas.read_sql`.

    This experimental feature provides distributed reading from a sql file.
    The function extended with `Spark-like parameters <https://spark.apache.org/docs/2.0.0/api/R/read.jdbc.html>`_
    such as ``partition_column``, ``lower_bound`` and ``upper_bound``. With these
    parameters, the user will be able to specify how to partition the imported data.

    Parameters
    ----------
    sql : str or SQLAlchemy Selectable (select or text object)
        SQL query to be executed or a table name.
    con : SQLAlchemy connectable, str, or sqlite3 connection
        Using SQLAlchemy makes it possible to use any DB supported by that
        library. If a DBAPI2 object, only sqlite3 is supported. The user is responsible
        for engine disposal and connection closure for the SQLAlchemy
        connectable; str connections are closed automatically. See
        `here <https://docs.sqlalchemy.org/en/13/core/connections.html>`_.
    index_col : str or list of str, optional
        Column(s) to set as index(MultiIndex).
    coerce_float : bool, default: True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point, useful for SQL result sets.
    params : list, tuple or dict, optional
        List of parameters to pass to execute method. The syntax used to pass
        parameters is database driver dependent. Check your database driver
        documentation for which of the five syntax styles, described in PEP 249's
        paramstyle, is supported. Eg. for psycopg2, uses %(name)s so use params=
        {'name' : 'value'}.
    parse_dates : list or dict, optional
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
          strftime compatible in case of parsing string times, or is one of
          (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
          to the keyword arguments of :func:`pandas.to_datetime`
          Especially useful with databases without native Datetime support,
          such as SQLite.
    columns : list, optional
        List of column names to select from SQL table (only used when reading
        a table).
    chunksize : int, optional
        If specified, return an iterator where `chunksize` is the
        number of rows to include in each chunk.
    dtype_backend : {"numpy_nullable", "pyarrow"}, default: NumPy backed DataFrames
        Which dtype_backend to use, e.g. whether a DataFrame should have NumPy arrays,
        nullable dtypes are used for all dtypes that have a nullable implementation when
        "numpy_nullable" is set, PyArrow is used for all dtypes if "pyarrow" is set.
        The dtype_backends are still experimential.
    dtype : Type name or dict of columns, optional
        Data type for data or columns. E.g. np.float64 or {'a': np.float64, 'b': np.int32, 'c': 'Int64'}. The argument is ignored if a table is passed instead of a query.
    partition_column : str, optional
        Column used to share the data between the workers (MUST be a INTEGER column).
    lower_bound : int, optional
        The minimum value to be requested from the partition_column.
    upper_bound : int, optional
        The maximum value to be requested from the partition_column.
    max_sessions : int, optional
        The maximum number of simultaneous connections allowed to use.

    Returns
    -------
    modin.DataFrame or Iterator[modin.DataFrame]
    """
    _, _, _, kwargs = inspect.getargvalues(inspect.currentframe())

    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    result = FactoryDispatcher.read_sql_distributed(**kwargs)
    if isinstance(result, BaseQueryCompiler):
        return DataFrame(query_compiler=result)
    return (DataFrame(query_compiler=qc) for qc in result)


@expanduser_path_arg("filepath_or_buffer")
def read_custom_text(
    filepath_or_buffer,
    columns,
    custom_parser,
    compression="infer",
    nrows: Optional[int] = None,
    is_quoting=True,
):
    r"""
    Load custom text data from file.

    Parameters
    ----------
    filepath_or_buffer : str
        File path where the custom text data will be loaded from.
    columns : list or callable(file-like object, \*\*kwargs) -> list
        Column names of list type or callable that create column names from opened file
        and passed `kwargs`.
    custom_parser : callable(file-like object, \*\*kwargs) -> pandas.DataFrame
        Function that takes as input a part of the `filepath_or_buffer` file loaded into
        memory in file-like object form.
    compression : {'infer', 'gzip', 'bz2', 'zip', 'xz', None}, default: 'infer'
        If 'infer' and 'path_or_url' is path-like, then detect compression from
        the following extensions: '.gz', '.bz2', '.zip', or '.xz' (otherwise no
        compression). If 'infer' and 'path_or_url' is not path-like, then use
        None (= no decompression).
    nrows : int, optional
        Amount of rows to read.
    is_quoting : bool, default: True
        Whether or not to consider quotes.

    Returns
    -------
    modin.DataFrame
    """
    _, _, _, kwargs = inspect.getargvalues(inspect.currentframe())

    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    return DataFrame(query_compiler=FactoryDispatcher.read_custom_text(**kwargs))


# CSV and table
def _make_parser_func(sep: str, funcname: str) -> Callable:
    """
    Create a parser function from the given sep.

    Parameters
    ----------
    sep : str
        The separator default to use for the parser.
    funcname : str
        The name of the generated parser function.

    Returns
    -------
    Callable
    """

    def parser_func(
        filepath_or_buffer: Union[str, pathlib.Path, IO[AnyStr]],
        *,
        sep=lib.no_default,
        delimiter=None,
        header="infer",
        names=lib.no_default,
        index_col=None,
        usecols=None,
        dtype=None,
        engine=None,
        converters=None,
        true_values=None,
        false_values=None,
        skipinitialspace=False,
        skiprows=None,
        skipfooter=0,
        nrows=None,
        na_values=None,
        keep_default_na=True,
        na_filter=True,
        verbose=lib.no_default,
        skip_blank_lines=True,
        parse_dates=None,
        infer_datetime_format=lib.no_default,
        keep_date_col=lib.no_default,
        date_parser=lib.no_default,
        date_format=None,
        dayfirst=False,
        cache_dates=True,
        iterator=False,
        chunksize=None,
        compression="infer",
        thousands=None,
        decimal: str = ".",
        lineterminator=None,
        quotechar='"',
        quoting=0,
        escapechar=None,
        comment=None,
        encoding=None,
        encoding_errors="strict",
        dialect=None,
        on_bad_lines="error",
        doublequote=True,
        delim_whitespace=lib.no_default,
        low_memory=True,
        memory_map=False,
        float_precision=None,
        storage_options: StorageOptions = None,
        dtype_backend=lib.no_default,
    ) -> DataFrame:
        # ISSUE #2408: parse parameter shared with pandas read_csv and read_table and update with provided args
        _pd_read_csv_signature = {
            val.name for val in inspect.signature(pandas.read_csv).parameters.values()
        }
        _, _, _, f_locals = inspect.getargvalues(inspect.currentframe())
        if f_locals.get("sep", sep) is False:
            f_locals["sep"] = "\t"

        kwargs = {k: v for k, v in f_locals.items() if k in _pd_read_csv_signature}
        return _read(**kwargs)

    parser_func.__doc__ = _read.__doc__
    parser_func.__name__ = funcname
    return expanduser_path_arg("filepath_or_buffer")(parser_func)


def _read(**kwargs) -> DataFrame:
    """
    General documentation is available in `modin.pandas.read_csv`.

    This experimental feature provides parallel reading from multiple csv files which are
    defined by glob pattern.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments in `modin.pandas.read_csv`.

    Returns
    -------
    modin.DataFrame

    Examples
    --------
    >>> import modin.experimental.pandas as pd
    >>> df = pd.read_csv_glob("s3://dask-data/nyc-taxi/2015/yellow_tripdata_2015-1*")
    UserWarning: `read_*` implementation has mismatches with pandas:
    Data types of partitions are different! Please refer to the troubleshooting section of the Modin documentation to fix this issue.
            VendorID tpep_pickup_datetime  ... total_amount  congestion_surcharge
    0             1.0  2020-10-01 00:09:08  ...         4.30                   0.0
    1             1.0  2020-10-01 00:09:19  ...        13.30                   2.5
    2             1.0  2020-10-01 00:30:00  ...        15.36                   2.5
    3             2.0  2020-10-01 00:56:46  ...        -3.80                   0.0
    4             2.0  2020-10-01 00:56:46  ...         3.80                   0.0
    ...           ...                  ...  ...          ...                   ...
    4652008       NaN  2020-12-31 23:44:35  ...        43.95                   2.5
    4652009       NaN  2020-12-31 23:41:36  ...        20.17                   2.5
    4652010       NaN  2020-12-31 23:01:17  ...        78.98                   0.0
    4652011       NaN  2020-12-31 23:31:29  ...        39.50                   0.0
    4652012       NaN  2020-12-31 23:12:48  ...        20.64                   0.0

    [4652013 rows x 18 columns]
    """
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    pd_obj = FactoryDispatcher.read_csv_glob(**kwargs)
    # This happens when `read_csv` returns a TextFileReader object for iterating through
    if isinstance(pd_obj, pandas.io.parsers.TextFileReader):
        reader = pd_obj.read
        pd_obj.read = lambda *args, **kwargs: DataFrame(
            query_compiler=reader(*args, **kwargs)
        )
        return pd_obj

    return DataFrame(query_compiler=pd_obj)


read_csv_glob = _make_parser_func(sep=",", funcname="read_csv_glob")


@expanduser_path_arg("filepath_or_buffer")
def read_pickle_glob(
    filepath_or_buffer,
    compression: Optional[str] = "infer",
    storage_options: StorageOptions = None,
):
    """
    Load pickled pandas object from files.

    This experimental feature provides parallel reading from multiple pickle files which are
    defined by glob pattern. The files must contain parts of one dataframe, which can be
    obtained, for example, by `DataFrame.modin.to_pickle_glob` function.

    Parameters
    ----------
    filepath_or_buffer : str, path object or file-like object
        File path, URL, or buffer where the pickled object will be loaded from.
        Accept URL. URL is not limited to S3 and GCS.
    compression : {{'infer', 'gzip', 'bz2', 'zip', 'xz', None}}, default: 'infer'
        If 'infer' and 'path_or_url' is path-like, then detect compression from
        the following extensions: '.gz', '.bz2', '.zip', or '.xz' (otherwise no
        compression) If 'infer' and 'path_or_url' is not path-like, then use
        None (= no decompression).
    storage_options : dict, optional
        Extra options that make sense for a particular storage connection, e.g.
        host, port, username, password, etc., if using a URL that will be parsed by
        fsspec, e.g., starting "s3://", "gcs://". An error will be raised if providing
        this argument with a non-fsspec URL. See the fsspec and backend storage
        implementation docs for the set of allowed keys and values.

    Returns
    -------
    unpickled : same type as object stored in file

    Notes
    -----
    The number of partitions is equal to the number of input files.
    """
    _, _, _, kwargs = inspect.getargvalues(inspect.currentframe())

    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    return DataFrame(query_compiler=FactoryDispatcher.read_pickle_glob(**kwargs))


@expanduser_path_arg("filepath_or_buffer")
def to_pickle_glob(
    self,
    filepath_or_buffer,
    compression: CompressionOptions = "infer",
    protocol: int = pickle.HIGHEST_PROTOCOL,
    storage_options: StorageOptions = None,
) -> None:
    """
    Pickle (serialize) object to file.

    This experimental feature provides parallel writing into multiple pickle files which are
    defined by glob pattern, otherwise (without glob pattern) default pandas implementation is used.

    Parameters
    ----------
    filepath_or_buffer : str
        File path where the pickled object will be stored.
    compression : {{'infer', 'gzip', 'bz2', 'zip', 'xz', None}}, default: 'infer'
        A string representing the compression to use in the output file. By
        default, infers from the file extension in specified path.
        Compression mode may be any of the following possible
        values: {{'infer', 'gzip', 'bz2', 'zip', 'xz', None}}. If compression
        mode is 'infer' and path_or_buf is path-like, then detect
        compression mode from the following extensions:
        '.gz', '.bz2', '.zip' or '.xz'. (otherwise no compression).
        If dict given and mode is 'zip' or inferred as 'zip', other entries
        passed as additional compression options.
    protocol : int, default: pickle.HIGHEST_PROTOCOL
        Int which indicates which protocol should be used by the pickler,
        default HIGHEST_PROTOCOL (see `pickle docs <https://docs.python.org/3/library/pickle.html>`_
        paragraph 12.1.2 for details). The possible  values are 0, 1, 2, 3, 4, 5. A negative value
        for the protocol parameter is equivalent to setting its value to HIGHEST_PROTOCOL.
    storage_options : dict, optional
        Extra options that make sense for a particular storage connection, e.g.
        host, port, username, password, etc., if using a URL that will be parsed by
        fsspec, e.g., starting "s3://", "gcs://". An error will be raised if providing
        this argument with a non-fsspec URL. See the fsspec and backend storage
        implementation docs for the set of allowed keys and values.
    """
    obj = self
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    if isinstance(self, DataFrame):
        obj = self._query_compiler
    FactoryDispatcher.to_pickle_glob(
        obj,
        filepath_or_buffer=filepath_or_buffer,
        compression=compression,
        protocol=protocol,
        storage_options=storage_options,
    )


@expanduser_path_arg("path")
def read_parquet_glob(
    path,
    engine: str = "auto",
    columns: list[str] | None = None,
    storage_options: StorageOptions = None,
    use_nullable_dtypes: bool = lib.no_default,
    dtype_backend=lib.no_default,
    filesystem=None,
    filters=None,
    **kwargs,
) -> DataFrame:  # noqa: PR01
    """
    Load a parquet object from the file path, returning a DataFrame.

    This experimental feature provides parallel reading from multiple parquet files which are
    defined by glob pattern. The files must contain parts of one dataframe, which can be
    obtained, for example, by `DataFrame.modin.to_parquet_glob` function.

    Returns
    -------
    DataFrame

    Notes
    -----
    * Only string type supported for `path` argument.
    * The rest of the arguments are the same as for `pandas.read_parquet`.
    """
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    return DataFrame(
        query_compiler=FactoryDispatcher.read_parquet_glob(
            path=path,
            engine=engine,
            columns=columns,
            storage_options=storage_options,
            use_nullable_dtypes=use_nullable_dtypes,
            dtype_backend=dtype_backend,
            filesystem=filesystem,
            filters=filters,
            **kwargs,
        )
    )


@expanduser_path_arg("path")
def to_parquet_glob(
    self,
    path,
    engine="auto",
    compression="snappy",
    index=None,
    partition_cols=None,
    storage_options: StorageOptions = None,
    **kwargs,
) -> None:  # noqa: PR01
    """
    Write a DataFrame to the binary parquet format.

    This experimental feature provides parallel writing into multiple parquet files which are
    defined by glob pattern, otherwise (without glob pattern) default pandas implementation is used.

    Notes
    -----
    * Only string type supported for `path` argument.
    * The rest of the arguments are the same as for `pandas.to_parquet`.
    """
    obj = self
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    if isinstance(self, DataFrame):
        obj = self._query_compiler
    FactoryDispatcher.to_parquet_glob(
        obj,
        path=path,
        engine=engine,
        compression=compression,
        index=index,
        partition_cols=partition_cols,
        storage_options=storage_options,
        **kwargs,
    )


@expanduser_path_arg("path_or_buf")
def read_json_glob(
    path_or_buf,
    *,
    orient: str | None = None,
    typ: Literal["frame", "series"] = "frame",
    dtype: DtypeArg | None = None,
    convert_axes=None,
    convert_dates: bool | list[str] = True,
    keep_default_dates: bool = True,
    precise_float: bool = False,
    date_unit: str | None = None,
    encoding: str | None = None,
    encoding_errors: str | None = "strict",
    lines: bool = False,
    chunksize: int | None = None,
    compression: CompressionOptions = "infer",
    nrows: int | None = None,
    storage_options: StorageOptions = None,
    dtype_backend: Union[DtypeBackend, lib.NoDefault] = lib.no_default,
    engine="ujson",
) -> DataFrame:  # noqa: PR01
    """
    Convert a JSON string to pandas object.

    This experimental feature provides parallel reading from multiple json files which are
    defined by glob pattern. The files must contain parts of one dataframe, which can be
    obtained, for example, by `DataFrame.modin.to_json_glob` function.

    Returns
    -------
    DataFrame

    Notes
    -----
    * Only string type supported for `path_or_buf` argument.
    * The rest of the arguments are the same as for `pandas.read_json`.
    """
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    if nrows is not None:
        raise NotImplementedError(
            "`read_json_glob` only support nrows is None, otherwise use `to_json`."
        )

    return DataFrame(
        query_compiler=FactoryDispatcher.read_json_glob(
            path_or_buf=path_or_buf,
            orient=orient,
            typ=typ,
            dtype=dtype,
            convert_axes=convert_axes,
            convert_dates=convert_dates,
            keep_default_dates=keep_default_dates,
            precise_float=precise_float,
            date_unit=date_unit,
            encoding=encoding,
            encoding_errors=encoding_errors,
            lines=lines,
            chunksize=chunksize,
            compression=compression,
            nrows=nrows,
            storage_options=storage_options,
            dtype_backend=dtype_backend,
            engine=engine,
        )
    )


@expanduser_path_arg("path_or_buf")
def to_json_glob(
    self,
    path_or_buf=None,
    orient=None,
    date_format=None,
    double_precision=10,
    force_ascii=True,
    date_unit="ms",
    default_handler=None,
    lines=False,
    compression="infer",
    index=None,
    indent=None,
    storage_options: StorageOptions = None,
    mode="w",
) -> None:  # noqa: PR01
    """
    Convert the object to a JSON string.

    Notes
    -----
    * Only string type supported for `path_or_buf` argument.
    * The rest of the arguments are the same as for `pandas.to_json`.
    """
    obj = self
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    if isinstance(self, DataFrame):
        obj = self._query_compiler
    FactoryDispatcher.to_json_glob(
        obj,
        path_or_buf=path_or_buf,
        orient=orient,
        date_format=date_format,
        double_precision=double_precision,
        force_ascii=force_ascii,
        date_unit=date_unit,
        default_handler=default_handler,
        lines=lines,
        compression=compression,
        index=index,
        indent=indent,
        storage_options=storage_options,
        mode=mode,
    )


@expanduser_path_arg("path_or_buffer")
def read_xml_glob(
    path_or_buffer,
    *,
    xpath="./*",
    namespaces=None,
    elems_only=False,
    attrs_only=False,
    names=None,
    dtype=None,
    converters=None,
    parse_dates=None,
    encoding="utf-8",
    parser="lxml",
    stylesheet=None,
    iterparse=None,
    compression="infer",
    storage_options: StorageOptions = None,
    dtype_backend=lib.no_default,
) -> DataFrame:  # noqa: PR01
    """
    Read XML document into a DataFrame object.

    This experimental feature provides parallel reading from multiple XML files which are
    defined by glob pattern. The files must contain parts of one dataframe, which can be
    obtained, for example, by `DataFrame.modin.to_xml_glob` function.

    Returns
    -------
    DataFrame

    Notes
    -----
    * Only string type supported for `path_or_buffer` argument.
    * The rest of the arguments are the same as for `pandas.read_xml`.
    """
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    return DataFrame(
        query_compiler=FactoryDispatcher.read_xml_glob(
            path_or_buffer=path_or_buffer,
            xpath=xpath,
            namespaces=namespaces,
            elems_only=elems_only,
            attrs_only=attrs_only,
            names=names,
            dtype=dtype,
            converters=converters,
            parse_dates=parse_dates,
            encoding=encoding,
            parser=parser,
            stylesheet=stylesheet,
            iterparse=iterparse,
            compression=compression,
            storage_options=storage_options,
            dtype_backend=dtype_backend,
        )
    )


@expanduser_path_arg("path_or_buffer")
def to_xml_glob(
    self,
    path_or_buffer=None,
    index=True,
    root_name="data",
    row_name="row",
    na_rep=None,
    attr_cols=None,
    elem_cols=None,
    namespaces=None,
    prefix=None,
    encoding="utf-8",
    xml_declaration=True,
    pretty_print=True,
    parser="lxml",
    stylesheet=None,
    compression="infer",
    storage_options=None,
) -> None:  # noqa: PR01
    """
    Render a DataFrame to an XML document.

    Notes
    -----
    * Only string type supported for `path_or_buffer` argument.
    * The rest of the arguments are the same as for `pandas.to_xml`.
    """
    obj = self
    from modin.core.execution.dispatching.factories.dispatcher import FactoryDispatcher

    if isinstance(self, DataFrame):
        obj = self._query_compiler
    FactoryDispatcher.to_xml_glob(
        obj,
        path_or_buffer=path_or_buffer,
        index=index,
        root_name=root_name,
        row_name=row_name,
        na_rep=na_rep,
        attr_cols=attr_cols,
        elem_cols=elem_cols,
        namespaces=namespaces,
        prefix=prefix,
        encoding=encoding,
        xml_declaration=xml_declaration,
        pretty_print=pretty_print,
        parser=parser,
        stylesheet=stylesheet,
        compression=compression,
        storage_options=storage_options,
    )



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Module with classes and utilities for deferred remote execution in Ray workers."""

from enum import Enum
from itertools import islice
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import pandas
import ray
from ray._private.services import get_node_ip_address
from ray.util.client.common import ClientObjectRef

from modin.core.execution.ray.common import MaterializationHook, RayWrapper
from modin.logging import get_logger

ObjectRefType = Union[ray.ObjectRef, ClientObjectRef, None]
ObjectRefOrListType = Union[ObjectRefType, List[ObjectRefType]]
ListOrTuple = (list, tuple)


class DeferredExecution:
    """
    Deferred execution task.

    This class represents a single node in the execution tree. The input is either
    an object reference or another node on which this node depends.
    The output is calculated by the specified Callable.

    If the input is a DeferredExecution node, it is executed first and the execution
    output is used as the input for this one. All the executions are performed in a
    single batch (i.e. using a single remote call) and the results are saved in all
    the nodes that have multiple subscribers.

    Parameters
    ----------
    data : ObjectRefType or DeferredExecution
        The execution input.
    func : callable or ObjectRefType
        A function to be executed.
    args : list or tuple
        Additional positional arguments to be passed in `func`.
    kwargs : dict
        Additional keyword arguments to be passed in `func`.
    num_returns : int, optional
        The number of the return values.

    Attributes
    ----------
    data : ObjectRefType or DeferredExecution
        The execution input.
    func : callable or ObjectRefType
        A function to be executed.
    args : list or tuple
        Additional positional arguments to be passed in `func`.
    kwargs : dict
        Additional keyword arguments to be passed in `func`.
    num_returns : int
        The number of the return values.
    flat_args : bool
        True means that there are no lists or DeferredExecution objects in `args`.
        In this case, no arguments processing is performed and `args` is passed
        to the remote method as is.
    flat_kwargs : bool
        The same as `flat_args` but for the `kwargs` values.
    """

    def __init__(
        self,
        data: Union[
            ObjectRefType,
            "DeferredExecution",
            List[Union[ObjectRefType, "DeferredExecution"]],
        ],
        func: Union[Callable, ObjectRefType],
        args: Union[List[Any], Tuple[Any]],
        kwargs: Dict[str, Any],
        num_returns=1,
    ):
        if isinstance(data, DeferredExecution):
            data.subscribe()
        self.data = data
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.num_returns = num_returns
        self.flat_args = self._flat_args(args)
        self.flat_kwargs = self._flat_args(kwargs.values())
        self.subscribers = 0

    @classmethod
    def _flat_args(cls, args: Iterable):
        """
        Check if the arguments list is flat and subscribe to all `DeferredExecution` objects.

        Parameters
        ----------
        args : Iterable

        Returns
        -------
        bool
        """
        flat = True
        for arg in args:
            if isinstance(arg, DeferredExecution):
                flat = False
                arg.subscribe()
            elif isinstance(arg, ListOrTuple):
                flat = False
                cls._flat_args(arg)
        return flat

    def exec(
        self,
    ) -> Tuple[ObjectRefOrListType, Union["MetaList", List], Union[int, List[int]]]:
        """
        Execute this task, if required.

        Returns
        -------
        tuple
            The execution result, MetaList, containing the length, width and
            the worker's ip address (the last value in the list) and the values
            offset in the list. I.e. length = meta_list[offset],
            width = meta_list[offset + 1], ip = meta_list[-1].
        """
        if self.has_result:
            return self.data, self.meta, self.meta_offset

        if (
            not isinstance(self.data, DeferredExecution)
            and self.flat_args
            and self.flat_kwargs
            and self.num_returns == 1
        ):
            result, length, width, ip = remote_exec_func.remote(
                self.func, self.data, *self.args, **self.kwargs
            )
            meta = MetaList([length, width, ip])
            self._set_result(result, meta, 0)
            return result, meta, 0

        # If there are no subscribers, we still need the result here. We don't need to decrement
        # it back. After the execution, the result is saved and the counter has no effect.
        self.subscribers += 2
        consumers, output = self._deconstruct()
        # The last result is the MetaList, so adding +1 here.
        num_returns = sum(c.num_returns for c in consumers) + 1
        results = self._remote_exec_chain(num_returns, *output)
        meta = MetaList(results.pop())
        meta_offset = 0
        results = iter(results)
        for de in consumers:
            if de.num_returns == 1:
                de._set_result(next(results), meta, meta_offset)
                meta_offset += 2
            else:
                res = list(islice(results, num_returns))
                offsets = list(range(0, 2 * num_returns, 2))
                de._set_result(res, meta, offsets)
                meta_offset += 2 * num_returns
        return self.data, self.meta, self.meta_offset

    @property
    def has_result(self):
        """
        Return true if this task has already been executed and the result is set.

        Returns
        -------
        bool
        """
        return not hasattr(self, "func")

    def subscribe(self):
        """
        Increment the `subscribers` counter.

        Subscriber is any instance that could trigger the execution of this task.
        In case of a multiple subscribers, the execution could be triggerred multiple
        times. To prevent the multiple executions, the execution result is returned
        from the worker and saved in this instance. Subsequent calls to `execute()`
        return the previously saved result.
        """
        self.subscribers += 1

    def unsubscribe(self):
        """Decrement the `subscribers` counter."""
        self.subscribers -= 1
        assert self.subscribers >= 0

    def _deconstruct(self) -> Tuple[List["DeferredExecution"], List[Any]]:
        """
        Convert the specified execution tree to a flat list.

        This is required for the automatic Ray object references
        materialization before passing the list to a Ray worker.

        The format of the list is the following:
        <input object> sequence<<function> <n><args> <n><kwargs> <ref> <nret>>...
        If <n> before <args> is >= 0, then the next n objects are the function arguments.
        If it is -1, it means that the method arguments contain list and/or
        DeferredExecution (chain) objects. In this case the next values are read
        one by one until `_Tag.END` is encountered. If the value is `_Tag.LIST`,
        then the next sequence of values up to `_Tag.END` is converted to list.
        If the value is `_Tag.CHAIN`, then the next sequence of values up to
        `_Tag.END` has exactly the same format, as described here.
        If the value is `_Tag.REF`, then the next value is a reference id, i.e.
        the actual value should be retrieved by this id from the previously
        saved objects. The <input object> could also be `_Tag.REF` or `_Tag.LIST`.

        If <n> before <kwargs> is >=0, then the next 2*n values are the argument
        names and values in the following format - [name1, value1, name2, value2...].
        If it's -1, then the next values are converted to list in the same way as
        <args> and the argument names are the next len(<args>) values.

        <ref> is an integer reference id. If it's not 0, then there is another
        chain referring to the execution result of this method and, thus, it must
        be saved so that other chains could retrieve the object by the id.

        <nret> field contains either the `num_returns` value or 0. If it's 0, the
        execution result is not returned, but is just passed to the next task in the
        chain. If it's 1, the result is returned as is. Otherwise, it's expected that
        the result is iterable and the specified number of values is returned from
        the iterator. The values lengths and widths are added to the meta list.

        Returns
        -------
        tuple of list
            * The first list is the result consumers.
                If a DeferredExecution has multiple subscribers, the execution result
                should be returned and saved in order to avoid duplicate executions.
                These DeferredExecution tasks are added to this list and, after the
                execution, the results are passed to the ``_set_result()`` method of
                each task.
            * The second is a flat list of arguments that could be passed to the remote executor.
        """
        stack = []
        result_consumers = []
        output = []
        # Using stack and generators to avoid the ``RecursionError``s.
        stack.append(self._deconstruct_chain(self, output, stack, result_consumers))
        while stack:
            try:
                gen = stack.pop()
                next_gen = next(gen)
                stack.append(gen)
                stack.append(next_gen)
            except StopIteration:
                pass
        return result_consumers, output

    @classmethod
    def _deconstruct_chain(
        cls,
        de: "DeferredExecution",
        output: List,
        stack: List,
        result_consumers: List["DeferredExecution"],
    ):
        """
        Deconstruct the specified DeferredExecution chain.

        Parameters
        ----------
        de : DeferredExecution
            The chain to be deconstructed.
        output : list
            Put the arguments to this list.
        stack : list
            Used to eliminate recursive calls, that may lead to the RecursionError.
        result_consumers : list of DeferredExecution
            The result consumers.

        Yields
        ------
        Generator
            The ``_deconstruct_list()`` generator.
        """
        out_append = output.append
        out_extend = output.extend
        while True:
            de.unsubscribe()
            if (out_pos := getattr(de, "out_pos", None)) and not de.has_result:
                out_append(_Tag.REF)
                out_append(out_pos)
                output[out_pos] = out_pos
                if de.subscribers == 0:
                    # We may have subscribed to the same node multiple times.
                    # It could happen, for example, if it's passed to the args
                    # multiple times, or it's one of the parent nodes and also
                    # passed to the args. In this case, there are no multiple
                    # subscribers, and we don't need to return the result.
                    output[out_pos + 1] = 0
                    result_consumers.remove(de)
                break
            elif not isinstance(data := de.data, DeferredExecution):
                if isinstance(data, ListOrTuple):
                    yield cls._deconstruct_list(
                        data, output, stack, result_consumers, out_append
                    )
                else:
                    out_append(data)
                if not de.has_result:
                    stack.append(de)
                break
            else:
                stack.append(de)
                de = data

        while stack and isinstance(stack[-1], DeferredExecution):
            de: DeferredExecution = stack.pop()
            args = de.args
            kwargs = de.kwargs
            out_append(de.func)
            if de.flat_args:
                out_append(len(args))
                out_extend(args)
            else:
                out_append(-1)
                yield cls._deconstruct_list(
                    args, output, stack, result_consumers, out_append
                )
            if de.flat_kwargs:
                out_append(len(kwargs))
                for item in kwargs.items():
                    out_extend(item)
            else:
                out_append(-1)
                yield cls._deconstruct_list(
                    kwargs.values(), output, stack, result_consumers, out_append
                )
                out_extend(kwargs)

            out_append(0)  # Placeholder for ref id
            if de.subscribers > 0:
                # Ref id. This is the index in the output list.
                de.out_pos = len(output) - 1
                result_consumers.append(de)
                out_append(de.num_returns)  # Return result for this node
            else:
                out_append(0)  # Do not return result for this node

    @classmethod
    def _deconstruct_list(
        cls,
        lst: Iterable,
        output: List,
        stack: List,
        result_consumers: List["DeferredExecution"],
        out_append: Callable,
    ):
        """
        Deconstruct the specified list.

        Parameters
        ----------
        lst : list
        output : list
        stack : list
        result_consumers : list
        out_append : Callable
            The reference to the ``list.append()`` method.

        Yields
        ------
        Generator
            Either ``_deconstruct_list()`` or ``_deconstruct_chain()`` generator.
        """
        for obj in lst:
            if isinstance(obj, DeferredExecution):
                if out_pos := getattr(obj, "out_pos", None):
                    obj.unsubscribe()
                    if obj.has_result:
                        out_append(obj.data)
                    else:
                        out_append(_Tag.REF)
                        out_append(out_pos)
                        output[out_pos] = out_pos
                        if obj.subscribers == 0:
                            output[out_pos + 1] = 0
                            result_consumers.remove(obj)
                else:
                    out_append(_Tag.CHAIN)
                    yield cls._deconstruct_chain(obj, output, stack, result_consumers)
                    out_append(_Tag.END)
            elif isinstance(obj, ListOrTuple):
                out_append(_Tag.LIST)
                yield cls._deconstruct_list(
                    obj, output, stack, result_consumers, out_append
                )
            else:
                out_append(obj)
        out_append(_Tag.END)

    @staticmethod
    def _remote_exec_chain(num_returns: int, *args: Tuple) -> List[Any]:
        """
        Execute the deconstructed chain in a worker process.

        Parameters
        ----------
        num_returns : int
            The number of return values.
        *args : tuple
            A deconstructed chain to be executed.

        Returns
        -------
        list
            The execution results. The last element of this list is the ``MetaList``.
        """
        # Prefer _remote_exec_single_chain(). It has fewer arguments and
        # does not require the num_returns to be specified in options.
        if num_returns == 2:
            return _remote_exec_single_chain.remote(*args)
        else:
            return _remote_exec_multi_chain.options(num_returns=num_returns).remote(
                num_returns, *args
            )

    def _set_result(
        self,
        result: ObjectRefOrListType,
        meta: "MetaList",
        meta_offset: Union[int, List[int]],
    ):
        """
        Set the execution result.

        Parameters
        ----------
        result : ObjectRefOrListType
        meta : MetaList
        meta_offset : int or list of int
        """
        del self.func, self.args, self.kwargs, self.flat_args, self.flat_kwargs
        self.data = result
        self.meta = meta
        self.meta_offset = meta_offset

    def __reduce__(self):
        """Not serializable."""
        raise NotImplementedError("DeferredExecution is not serializable!")


class MetaList:
    """
    Meta information, containing the result lengths and the worker address.

    Parameters
    ----------
    obj : ray.ObjectID or list
    """

    def __init__(self, obj: Union[ray.ObjectID, ClientObjectRef, List]):
        self._obj = obj

    def __getitem__(self, index):
        """
        Get item at the specified index.

        Parameters
        ----------
        index : int

        Returns
        -------
        Any
        """
        obj = self._obj
        return obj[index] if isinstance(obj, list) else MetaListHook(self, index)

    def __setitem__(self, index, value):
        """
        Set item at the specified index.

        Parameters
        ----------
        index : int
        value : Any
        """
        obj = self._obj
        if not isinstance(obj, list):
            self._obj = obj = RayWrapper.materialize(obj)
        obj[index] = value


class MetaListHook(MaterializationHook):
    """
    Used by MetaList.__getitem__() for lazy materialization and getting a single value from the list.

    Parameters
    ----------
    meta : MetaList
        Non-materialized list to get the value from.
    idx : int
        The value index in the list.
    """

    def __init__(self, meta: MetaList, idx: int):
        self.meta = meta
        self.idx = idx

    def pre_materialize(self):
        """
        Get item at self.idx or object ref if not materialized.

        Returns
        -------
        object
        """
        obj = self.meta._obj
        return obj[self.idx] if isinstance(obj, list) else obj

    def post_materialize(self, materialized):
        """
        Save the materialized list in self.meta and get the item at self.idx.

        Parameters
        ----------
        materialized : list

        Returns
        -------
        object
        """
        self.meta._obj = materialized
        return materialized[self.idx]


class _Tag(Enum):  # noqa: PR01
    """
    A set of special values used for the method arguments de/construction.

    See ``DeferredExecution._deconstruct()`` for details.
    """

    # The next item is an execution chain
    CHAIN = 0
    # The next item is a reference
    REF = 1
    # The next item a list
    LIST = 2
    # End of list or chain
    END = 3


class _RemoteExecutor:
    """Remote functions for DeferredExecution."""

    @staticmethod
    def exec_func(fn: Callable, obj: Any, args: Tuple, kwargs: Dict) -> Any:
        """
        Execute the specified function.

        Parameters
        ----------
        fn : Callable
        obj : Any
        args : Tuple
        kwargs : dict

        Returns
        -------
        Any
        """
        try:
            try:
                return fn(obj, *args, **kwargs)
                # Sometimes Arrow forces us to make a copy of an object before we operate on it. We
                # don't want the error to propagate to the user, and we want to avoid copying unless
                # we absolutely have to.
            except ValueError as err:
                if isinstance(obj, (pandas.DataFrame, pandas.Series)):
                    return fn(obj.copy(), *args, **kwargs)
                else:
                    raise err
        except Exception as err:
            get_logger().error(
                f"{err}. fn={fn}, obj={obj}, args={args}, kwargs={kwargs}"
            )
            raise err

    @classmethod
    def construct(cls, num_returns: int, args: Tuple):  # pragma: no cover
        """
        Construct and execute the specified chain.

        This function is called in a worker process. The last value, returned by
        this generator, is the meta list, containing the objects lengths and widths
        and the worker ip address, as the last value in the list.

        Parameters
        ----------
        num_returns : int
        args : tuple

        Yields
        ------
        Any
            The execution results and the MetaList as the last value.
        """
        chain = list(reversed(args))
        meta = []
        try:
            stack = [cls.construct_chain(chain, {}, meta, None)]
            while stack:
                try:
                    gen = stack.pop()
                    obj = next(gen)
                    stack.append(gen)
                    if isinstance(obj, Generator):
                        stack.append(obj)
                    else:
                        yield obj
                except StopIteration:
                    pass
        except Exception as err:
            get_logger().error(f"{err}. args={args}, chain={list(reversed(chain))}")
            raise err
        meta.append(get_node_ip_address())
        yield meta

    @classmethod
    def construct_chain(
        cls,
        chain: List,
        refs: Dict[int, Any],
        meta: List,
        lst: Optional[List],
    ):  # pragma: no cover
        """
        Construct the chain and execute it one by one.

        Parameters
        ----------
        chain : list
            A flat list containing the execution tree, deconstructed by
            ``DeferredExecution._deconstruct()``.
        refs : dict
            If an execution result is required for multiple chains, the
            reference to this result is saved in this dict.
        meta : list
            The lengths of the returned objects are added to this list.
        lst : list
            If specified, the execution result is added to this list.
            This is used when a chain is passed as an argument to a
            DeferredExecution task.

        Yields
        ------
        Any
            Either the ``construct_list()`` generator or the execution results.
        """
        pop = chain.pop
        tg_e = _Tag.END

        obj = pop()
        if obj is _Tag.REF:
            obj = refs[pop()]
        elif obj is _Tag.LIST:
            obj = []
            yield cls.construct_list(obj, chain, refs, meta)

        while chain:
            fn = pop()
            if fn == tg_e:
                lst.append(obj)
                break

            if (args_len := pop()) >= 0:
                if args_len == 0:
                    args = []
                else:
                    args = chain[-args_len:]
                    del chain[-args_len:]
                    args.reverse()
            else:
                args = []
                yield cls.construct_list(args, chain, refs, meta)
            if (args_len := pop()) >= 0:
                kwargs = {pop(): pop() for _ in range(args_len)}
            else:
                values = []
                yield cls.construct_list(values, chain, refs, meta)
                kwargs = {pop(): v for v in values}

            obj = cls.exec_func(fn, obj, args, kwargs)

            if ref := pop():  # <ref> is not 0 - adding the result to refs
                refs[ref] = obj
            if (num_returns := pop()) == 0:
                continue

            itr = iter([obj] if num_returns == 1 else obj)
            for _ in range(num_returns):
                obj = next(itr)
                meta.append(len(obj) if hasattr(obj, "__len__") else 0)
                meta.append(len(obj.columns) if hasattr(obj, "columns") else 0)
                yield obj

    @classmethod
    def construct_list(
        cls,
        lst: List,
        chain: List,
        refs: Dict[int, Any],
        meta: List,
    ):  # pragma: no cover
        """
        Construct the list.

        Parameters
        ----------
        lst : list
        chain : list
        refs : dict
        meta : list

        Yields
        ------
        Any
            Either ``construct_chain()`` or ``construct_list()`` generator.
        """
        pop = chain.pop
        lst_append = lst.append
        while True:
            obj = pop()
            if isinstance(obj, _Tag):
                if obj == _Tag.END:
                    break
                elif obj == _Tag.CHAIN:
                    yield cls.construct_chain(chain, refs, meta, lst)
                elif obj == _Tag.LIST:
                    lst_append([])
                    yield cls.construct_list(lst[-1], chain, refs, meta)
                elif obj is _Tag.REF:
                    lst_append(refs[pop()])
                else:
                    raise ValueError(f"Unexpected tag {obj}")
            else:
                lst_append(obj)

    def __reduce__(self):
        """
        Use a single instance on deserialization.

        Returns
        -------
        str
            Returns the ``_REMOTE_EXEC`` attribute name.
        """
        return "_REMOTE_EXEC"


_REMOTE_EXEC = _RemoteExecutor()


@ray.remote(num_returns=4)
def remote_exec_func(
    fn: Callable,
    obj: Any,
    *flat_args: Tuple,
    remote_executor=_REMOTE_EXEC,
    **flat_kwargs: Dict,
):  # pragma: no cover
    """
    Execute the specified function with the arguments in a worker process.

    The object `obj` is passed to the function as the first argument.
    Note: all the arguments must be flat, i.e. no lists, no chains.

    Parameters
    ----------
    fn : Callable
    obj : Any
    *flat_args : list
    remote_executor : _RemoteExecutor, default: _REMOTE_EXEC
        Do not change, it's used to avoid excessive serializations.
    **flat_kwargs : dict

    Returns
    -------
    tuple[Any, int, int, str]
    The execution result, the result length and width, the worked address.
    """
    obj = remote_executor.exec_func(fn, obj, flat_args, flat_kwargs)
    return (
        obj,
        len(obj) if hasattr(obj, "__len__") else 0,
        len(obj.columns) if hasattr(obj, "columns") else 0,
        get_node_ip_address(),
    )


@ray.remote(num_returns=2)
def _remote_exec_single_chain(
    *args: Tuple, remote_executor=_REMOTE_EXEC
) -> Generator:  # pragma: no cover
    """
    Execute the deconstructed chain with a single return value in a worker process.

    Parameters
    ----------
    *args : tuple
        A deconstructed chain to be executed.
    remote_executor : _RemoteExecutor, default: _REMOTE_EXEC
        Do not change, it's used to avoid excessive serializations.

    Returns
    -------
    Generator
    """
    return remote_executor.construct(num_returns=2, args=args)


@ray.remote
def _remote_exec_multi_chain(
    num_returns: int, *args: Tuple, remote_executor=_REMOTE_EXEC
) -> Generator:  # pragma: no cover
    """
    Execute the deconstructed chain with a multiple return values in a worker process.

    Parameters
    ----------
    num_returns : int
        The number of return values.
    *args : tuple
        A deconstructed chain to be executed.
    remote_executor : _RemoteExecutor, default: _REMOTE_EXEC
        Do not change, it's used to avoid excessive serializations.

    Returns
    -------
    Generator
    """
    return remote_executor.construct(num_returns, args)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import numpy

from modin.error_message import ErrorMessage
from modin.utils import _inherit_docstrings

from .arr import array
from .utils import try_convert_from_interoperable_type


def _dispatch_math(operator_name, arr_method_name=None):
    # `operator_name` is the name of the method on the numpy API
    # `arr_method_name` is the name of the method on the modin.numpy.array object,
    # which is assumed to be `operator_name` by default
    @_inherit_docstrings(getattr(numpy, operator_name))
    def call(x, *args, **kwargs):
        x = try_convert_from_interoperable_type(x)
        if not isinstance(x, array):
            ErrorMessage.bad_type_for_numpy_op(operator_name, type(x))
            return getattr(numpy, operator_name)(x, *args, **kwargs)

        return getattr(x, arr_method_name or operator_name)(*args, **kwargs)

    return call


absolute = _dispatch_math("absolute")
abs = absolute
add = _dispatch_math("add", "__add__")
divide = _dispatch_math("divide")
dot = _dispatch_math("dot")
float_power = _dispatch_math("float_power")
floor_divide = _dispatch_math("floor_divide")
power = _dispatch_math("power")
prod = _dispatch_math("prod")
multiply = _dispatch_math("multiply")
remainder = _dispatch_math("remainder")
mod = remainder
subtract = _dispatch_math("subtract")
sum = _dispatch_math("sum")
true_divide = _dispatch_math("true_divide", "divide")
mean = _dispatch_math("mean")


def var(x1, axis=None, dtype=None, out=None, keepdims=None, *, where=True):
    x1 = try_convert_from_interoperable_type(x1)
    if not isinstance(x1, array):
        ErrorMessage.bad_type_for_numpy_op("var", type(x1))
        return numpy.var(
            x1, axis=axis, out=out, keepdims=keepdims, where=where, dtype=dtype
        )
    return x1.var(axis=axis, out=out, keepdims=keepdims, where=where, dtype=dtype)


# Maximum and minimum are ufunc's in NumPy, which means that our array's __array_ufunc__
# implementation will automatically handle this. We still need the function though, so that
# if the operands are modin.pandas objects, we can convert them to arrays, but after that
# we can just use NumPy's maximum/minimum since that will route to our array's ufunc.
def maximum(
    x1, x2, out=None, where=True, casting="same_kind", order="K", dtype=None, subok=True
):
    x1 = try_convert_from_interoperable_type(x1)
    if not isinstance(x1, array):
        ErrorMessage.bad_type_for_numpy_op("maximum", type(x1))
    return numpy.maximum(
        x1,
        x2,
        out=out,
        where=where,
        casting=casting,
        order=order,
        dtype=dtype,
        subok=subok,
    )


def minimum(
    x1, x2, out=None, where=True, casting="same_kind", order="K", dtype=None, subok=True
):
    x1 = try_convert_from_interoperable_type(x1)
    if not isinstance(x1, array):
        ErrorMessage.bad_type_for_numpy_op("minimum", type(x1))
    return numpy.minimum(
        x1,
        x2,
        out=out,
        where=where,
        casting=casting,
        order=order,
        dtype=dtype,
        subok=subok,
    )


amax = _dispatch_math("amax", "max")
amin = _dispatch_math("amin", "min")
max = amax
min = amin


def sqrt(
    x, out=None, *, where=True, casting="same_kind", order="K", dtype=None, subok=True
):
    x = try_convert_from_interoperable_type(x)
    if not isinstance(x, array):
        ErrorMessage.bad_type_for_numpy_op("sqrt", type(x))
        return numpy.sqrt(
            x,
            out=out,
            where=where,
            casting=casting,
            order=order,
            dtype=dtype,
            subok=subok,
        )
    return x.sqrt(out, where, casting, order, dtype, subok)


def exp(
    x, out=None, *, where=True, casting="same_kind", order="K", dtype=None, subok=True
):
    x = try_convert_from_interoperable_type(x)
    if not isinstance(x, array):
        ErrorMessage.bad_type_for_numpy_op("exp", type(x))
        return numpy.exp(
            x,
            out=out,
            where=where,
            casting=casting,
            order=order,
            dtype=dtype,
            subok=subok,
        )
    return x.exp(out, where, casting, order, dtype, subok)


def argmax(a, axis=None, out=None, *, keepdims=None):
    a = try_convert_from_interoperable_type(a)
    if not isinstance(a, array):
        ErrorMessage.bad_type_for_numpy_op("argmax", type(a))
        return numpy.argmax(a, axis=axis, out=out, keepdims=keepdims)
    return a.argmax(axis=axis, out=out, keepdims=keepdims)


def argmin(a, axis=None, out=None, *, keepdims=None):
    a = try_convert_from_interoperable_type(a)
    if not isinstance(a, array):
        ErrorMessage.bad_type_for_numpy_op("argmin", type(a))
        return numpy.argmin(a, axis=axis, out=out, keepdims=keepdims)
    return a.argmin(axis=axis, out=out, keepdims=keepdims)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Module contains the functions designed for the enable/disable of logging.

``enable_logging`` is used for decorating individual Modin functions or classes.
"""

from functools import wraps
from types import FunctionType, MethodType
from typing import Any, Callable, Dict, Optional, Tuple, Type, Union

from modin.config import LogMode

from .config import LogLevel, get_logger

_MODIN_LOGGER_NOWRAP = "__modin_logging_nowrap__"


def disable_logging(func: Callable) -> Callable:
    """
    Disable logging of one particular function. Useful for decorated classes.

    Parameters
    ----------
    func : callable
        A method in a logger-decorated class for which logging should be disabled.

    Returns
    -------
    func
        A function with logging disabled.
    """
    setattr(func, _MODIN_LOGGER_NOWRAP, True)
    return func


def enable_logging(
    modin_layer: Union[str, Callable, Type] = "PANDAS-API",
    name: Optional[str] = None,
    log_level: LogLevel = LogLevel.INFO,
) -> Callable:
    """
    Log Decorator used on specific Modin functions or classes.

    Parameters
    ----------
    modin_layer : str or object to decorate, default: "PANDAS-API"
        Specified by the logger (e.g. PANDAS-API).
        If it's an object to decorate, call logger_decorator() on it with default arguments.
    name : str, optional
        The name of the object the decorator is being applied to.
        Composed from the decorated object name if not specified.
    log_level : LogLevel, default: LogLevel.INFO
        The log level (LogLevel.INFO, LogLevel.DEBUG, LogLevel.WARNING, etc.).

    Returns
    -------
    func
        A decorator function.
    """
    if not isinstance(modin_layer, str):
        # assume the decorator is used in a form without parenthesis like:
        # @enable_logging
        # def func()
        return enable_logging()(modin_layer)

    def decorator(obj: Any) -> Any:
        """Decorate function or class to add logs to Modin API function(s)."""
        if isinstance(obj, type):
            seen: Dict[Any, Any] = {}
            for attr_name, attr_value in vars(obj).items():
                if isinstance(
                    attr_value, (FunctionType, MethodType, classmethod, staticmethod)
                ) and not hasattr(attr_value, _MODIN_LOGGER_NOWRAP):
                    try:
                        wrapped = seen[attr_value]
                    except KeyError:
                        wrapped = seen[attr_value] = enable_logging(
                            modin_layer,
                            f"{name or obj.__name__}.{attr_name}",
                            log_level,
                        )(attr_value)

                    setattr(obj, attr_name, wrapped)
            return obj
        elif isinstance(obj, classmethod):
            return classmethod(decorator(obj.__func__))
        elif isinstance(obj, staticmethod):
            return staticmethod(decorator(obj.__func__))

        assert isinstance(modin_layer, str), "modin_layer is somehow not a string!"

        start_line = f"START::{modin_layer.upper()}::{name or obj.__name__}"
        stop_line = f"STOP::{modin_layer.upper()}::{name or obj.__name__}"

        @wraps(obj)
        def run_and_log(*args: Tuple, **kwargs: Dict) -> Any:
            """
            Compute function with logging if Modin logging is enabled.

            Parameters
            ----------
            *args : tuple
                The function arguments.
            **kwargs : dict
                The function keyword arguments.

            Returns
            -------
            Any
            """
            if LogMode.get() == "disable":
                return obj(*args, **kwargs)

            logger = get_logger()
            logger.log(log_level, start_line)
            try:
                result = obj(*args, **kwargs)
            except BaseException as e:
                # Only log the exception if a deeper layer of the modin stack has not
                # already logged it.
                if not hasattr(e, "_modin_logged"):
                    # use stack_info=True so that even if we are a few layers deep in
                    # modin, we log a stack trace that includes calls to higher layers
                    # of modin
                    get_logger("modin.logger.errors").exception(
                        stop_line, stack_info=True
                    )
                    e._modin_logged = True  # type: ignore[attr-defined]
                raise
            finally:
                logger.log(log_level, stop_line)
            return result

        # make sure we won't decorate multiple times
        return disable_logging(run_and_log)

    return decorator



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import glob
import os
import shutil
import uuid

from fuzzydata.clients.modin import ModinWorkflow
from fuzzydata.core.generator import generate_workflow

from modin.config import Engine


def test_fuzzydata_sample_workflow():
    # Workflow Generation Options
    wf_name = str(uuid.uuid4())[:8]  # Unique name for the generated workflow
    num_versions = 10  # Number of unique CSV files to generate
    cols = 33  # Columns in Base Artifact
    rows = 1000  # Rows in Base Artifact
    bfactor = 1.0  # Branching Factor - 0.1 is linear, 10.0 is star-like
    exclude_ops = ["groupby"]  # In-Memory groupby operations cause issue #4287
    matfreq = 2  # How many operations to chain before materialization

    engine = Engine.get().lower()

    # Create Output Directory for Workflow Data
    base_out_directory = (
        f"/tmp/fuzzydata-test-wf-{engine}/"  # Must match corresponding github-action
    )
    if os.path.exists(base_out_directory):
        shutil.rmtree(base_out_directory)
    output_directory = f"{base_out_directory}/{wf_name}/"
    os.makedirs(output_directory, exist_ok=True)

    # Start Workflow Generation
    workflow = generate_workflow(
        workflow_class=ModinWorkflow,
        name=wf_name,
        num_versions=num_versions,
        base_shape=(cols, rows),
        out_directory=output_directory,
        bfactor=bfactor,
        exclude_ops=exclude_ops,
        matfreq=matfreq,
        wf_options={"modin_engine": engine},
    )

    # Assertions that the workflow generation worked correctly
    assert len(workflow) == num_versions
    assert len(list(glob.glob(f"{output_directory}/artifacts/*.csv"))) == len(
        workflow.artifact_dict
    )
    assert os.path.exists(f"{output_directory}/{workflow.name}_operations.json")
    assert os.path.getsize(f"{output_directory}/{workflow.name}_operations.json") > 0
    assert os.path.exists(f"{output_directory}/{workflow.name}_gt_graph.csv")



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import modin.pandas as pd
from modin.pandas.api.extensions import register_dataframe_accessor


def test_dataframe_extension_simple_method():
    expected_string_val = "Some string value"
    method_name = "new_method"
    df = pd.DataFrame([1, 2, 3])

    @register_dataframe_accessor(method_name)
    def my_method_implementation(self):
        return expected_string_val

    assert method_name in pd.dataframe._DATAFRAME_EXTENSIONS_.keys()
    assert pd.dataframe._DATAFRAME_EXTENSIONS_[method_name] is my_method_implementation
    assert df.new_method() == expected_string_val


def test_dataframe_extension_non_method():
    expected_val = 4
    attribute_name = "four"
    register_dataframe_accessor(attribute_name)(expected_val)
    df = pd.DataFrame([1, 2, 3])

    assert attribute_name in pd.dataframe._DATAFRAME_EXTENSIONS_.keys()
    assert pd.dataframe._DATAFRAME_EXTENSIONS_[attribute_name] == 4
    assert df.four == expected_val


def test_dataframe_extension_accessing_existing_methods():
    df = pd.DataFrame([1, 2, 3])
    method_name = "self_accessor"
    expected_result = df.sum() / df.count()

    @register_dataframe_accessor(method_name)
    def my_average(self):
        return self.sum() / self.count()

    assert method_name in pd.dataframe._DATAFRAME_EXTENSIONS_.keys()
    assert pd.dataframe._DATAFRAME_EXTENSIONS_[method_name] is my_average
    assert df.self_accessor().equals(expected_result)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
from __future__ import annotations

from types import ModuleType
from typing import Type

import numpy as np
import pandas
import pytest
import ray
import torch
from torch.utils.data import RandomSampler, Sampler, SequentialSampler

import modin.pandas as pd
from modin.experimental.torch.datasets import ModinDataLoader


@pytest.fixture(scope="module", autouse=True)
def ray_fix():
    ray.init(num_cpus=1)
    yield None
    ray.shutdown()


def _load_test_dataframe(lib: ModuleType):
    df = lib.read_csv(
        "https://raw.githubusercontent.com/ponder-org/ponder-datasets/main/USA_Housing.csv"
    )
    return df


@pytest.mark.parametrize("lib", [pandas, pd])
@pytest.mark.parametrize("sampler_cls", [RandomSampler, SequentialSampler])
@pytest.mark.parametrize("batch_size", [16, 37])
def test_torch_dataloader(lib: ModuleType, sampler_cls: Type[Sampler], batch_size: int):
    df = _load_test_dataframe(lib)
    np.random.seed(42)
    torch.manual_seed(42)
    loader = ModinDataLoader(
        df,
        batch_size=batch_size,
        features=[
            "AVG_AREA_INCOME",
            "AVG_AREA_HOUSE_AGE",
            "AVG_AREA_NUM_ROOMS",
            "AVG_AREA_NUM_BEDROOMS",
            "POPULATION",
            "PRICE",
        ],
        sampler=sampler_cls,
    )

    outputs = []
    for batch in loader:
        assert batch.shape[0] <= batch_size, batch.shape
        assert batch.shape[1] == 6, batch.shape

        outputs.append(batch)

    return outputs


@pytest.mark.parametrize("sampler_cls", [RandomSampler, SequentialSampler])
@pytest.mark.parametrize("batch_size", [16, 37])
def test_compare_dataloaders(sampler_cls: Type[Sampler], batch_size: int):
    by_modin = test_torch_dataloader(pd, sampler_cls, batch_size=batch_size)
    by_pandas = test_torch_dataloader(pandas, sampler_cls, batch_size=batch_size)

    assert len(by_modin) == len(by_pandas)
    for tensor_by_modin, tensor_by_pandas in zip(by_modin, by_pandas):
        assert np.allclose(tensor_by_modin, tensor_by_pandas), (
            tensor_by_modin - tensor_by_pandas
        )



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import json
from textwrap import dedent, indent
from unittest.mock import Mock, patch

import numpy as np
import pandas
import pytest

import modin.pandas as pd
import modin.utils
from modin.error_message import ErrorMessage
from modin.pandas.test.utils import create_test_dfs


# Note: classes below are used for purely testing purposes - they
# simulate real-world use cases for _inherit_docstring
class BaseParent:
    def method(self):
        """ordinary method (base)"""

    def base_method(self):
        """ordinary method in base only"""

    @property
    def prop(self):
        """property"""

    @staticmethod
    def static():
        """static method"""

    @classmethod
    def clsmtd(cls):
        """class method"""


class BaseChild(BaseParent):
    """this is class docstring"""

    def method(self):
        """ordinary method (child)"""

    def own_method(self):
        """own method"""

    def no_overwrite(self):
        """another own method"""

    F = property(method)


@pytest.fixture(scope="module")
def wrapped_cls():
    @modin.utils._inherit_docstrings(BaseChild)
    class Wrapped:
        def method(self):
            pass

        def base_method(self):
            pass

        def own_method(self):
            pass

        def no_overwrite(self):
            """not overwritten doc"""

        @property
        def prop(self):
            return None

        @staticmethod
        def static():
            pass

        @classmethod
        def clsmtd(cls):
            pass

        F = property(method)

    return Wrapped


def _check_doc(wrapped, orig):
    assert wrapped.__doc__ == orig.__doc__
    if isinstance(wrapped, property):
        assert wrapped.fget.__doc_inherited__
    else:
        assert wrapped.__doc_inherited__


def test_doc_inherit_clslevel(wrapped_cls):
    _check_doc(wrapped_cls, BaseChild)


def test_doc_inherit_methods(wrapped_cls):
    _check_doc(wrapped_cls.method, BaseChild.method)
    _check_doc(wrapped_cls.base_method, BaseParent.base_method)
    _check_doc(wrapped_cls.own_method, BaseChild.own_method)
    assert wrapped_cls.no_overwrite.__doc__ != BaseChild.no_overwrite.__doc__
    assert not getattr(wrapped_cls.no_overwrite, "__doc_inherited__", False)


def test_doc_inherit_special(wrapped_cls):
    _check_doc(wrapped_cls.static, BaseChild.static)
    _check_doc(wrapped_cls.clsmtd, BaseChild.clsmtd)


def test_doc_inherit_props(wrapped_cls):
    assert type(wrapped_cls.method) == type(BaseChild.method)  # noqa: E721
    _check_doc(wrapped_cls.prop, BaseChild.prop)
    _check_doc(wrapped_cls.F, BaseChild.F)


def test_doc_inherit_prop_builder():
    def builder(name):
        return property(lambda self: name)

    class Parent:
        prop = builder("Parent")

    @modin.utils._inherit_docstrings(Parent)
    class Child(Parent):
        prop = builder("Child")

    assert Parent().prop == "Parent"
    assert Child().prop == "Child"


@pytest.mark.parametrize(
    "source_doc,to_append,expected",
    [
        (
            "One-line doc.",
            "One-line message.",
            "One-line doc.One-line message.",
        ),
        (
            """
            Regular doc-string
                With the setted indent style.
            """,
            """
                    Doc-string having different indents
                        in comparison with the regular one.
            """,
            """
            Regular doc-string
                With the setted indent style.

            Doc-string having different indents
                in comparison with the regular one.
            """,
        ),
    ],
)
def test_append_to_docstring(source_doc, to_append, expected):
    def source_fn():
        pass

    source_fn.__doc__ = source_doc
    result_fn = modin.utils.append_to_docstring(to_append)(source_fn)

    answer = dedent(result_fn.__doc__)
    expected = dedent(expected)

    assert answer == expected


def test_align_indents():
    source = """
    Source string that sets
        the indent pattern."""
    target = indent(source, " " * 5)
    result = modin.utils.align_indents(source, target)
    assert source == result


def test_format_string():
    template = """
            Source template string that has some {inline_placeholder}s.
            Placeholder1:
            {new_line_placeholder1}
            Placeholder2:
            {new_line_placeholder2}
            Placeholder3:
            {new_line_placeholder3}
            Placeholder4:
            {new_line_placeholder4}Text text:
                Placeholder5:
                {new_line_placeholder5}
    """

    singleline_value = "Single-line value"
    multiline_value = """
        Some string
            Having different indentation
        From the source one."""
    multiline_value_new_line_at_the_end = multiline_value + "\n"
    multiline_value_new_line_at_the_begin = "\n" + multiline_value

    expected = """
            Source template string that has some Single-line values.
            Placeholder1:
            Some string
                Having different indentation
            From the source one.
            Placeholder2:
            Single-line value
            Placeholder3:
            
            Some string
                Having different indentation
            From the source one.
            Placeholder4:
            Some string
                Having different indentation
            From the source one.
            Text text:
                Placeholder5:
                Some string
                    Having different indentation
                From the source one.
    """  # noqa: W293
    answer = modin.utils.format_string(
        template,
        inline_placeholder=singleline_value,
        new_line_placeholder1=multiline_value,
        new_line_placeholder2=singleline_value,
        new_line_placeholder3=multiline_value_new_line_at_the_begin,
        new_line_placeholder4=multiline_value_new_line_at_the_end,
        new_line_placeholder5=multiline_value,
    )
    assert answer == expected


def warns_that_defaulting_to_pandas(prefix=None, suffix=None):
    """
    Assert that code warns that it's defaulting to pandas.

    Parameters
    ----------
    prefix : Optional[str]
        If specified, checks that the start of the warning message matches this argument
        before "[Dd]efaulting to pandas".
    suffix : Optional[str]
        If specified, checks that the end of the warning message matches this argument
        after "[Dd]efaulting to pandas".

    Returns
    -------
    pytest.recwarn.WarningsChecker
        A WarningsChecker checking for a UserWarning saying that Modin is
        defaulting to Pandas.
    """
    match = "[Dd]efaulting to pandas"
    if prefix:
        # Message may be separated by newlines
        match = match + "(.|\\n)+"
    if suffix:
        match += "(.|\\n)+" + suffix
    return pytest.warns(UserWarning, match=match)


@pytest.mark.parametrize("as_json", [True, False])
def test_show_versions(as_json, capsys):
    modin.utils.show_versions(as_json=as_json)
    versions = capsys.readouterr().out
    assert modin.__version__ in versions

    if as_json:
        versions = json.loads(versions)
        assert versions["modin dependencies"]["modin"] == modin.__version__


def test_warns_that_defaulting_to_pandas():
    with warns_that_defaulting_to_pandas():
        ErrorMessage.default_to_pandas()

    with warns_that_defaulting_to_pandas():
        ErrorMessage.default_to_pandas(message="Function name")


def test_assert_dtypes_equal():
    """Verify that `assert_dtypes_equal` from test utils works correctly (raises an error when it has to)."""
    from modin.pandas.test.utils import assert_dtypes_equal

    # Serieses with equal dtypes
    sr1, sr2 = pd.Series([1.0]), pandas.Series([1.0])
    assert sr1.dtype == sr2.dtype == "float"
    assert_dtypes_equal(sr1, sr2)  # shouldn't raise an error since dtypes are equal

    # Serieses with different dtypes belonging to the same class
    sr1 = sr1.astype("int")
    assert sr1.dtype != sr2.dtype and sr1.dtype == "int"
    assert_dtypes_equal(sr1, sr2)  # shouldn't raise an error since both are numeric

    # Serieses with different dtypes not belonging to the same class
    sr2 = sr2.astype("str")
    assert sr1.dtype != sr2.dtype and sr2.dtype == "object"
    with pytest.raises(AssertionError):
        assert_dtypes_equal(sr1, sr2)

    # Dfs with equal dtypes
    df1, df2 = create_test_dfs({"a": [1], "b": [1.0]})
    assert_dtypes_equal(df1, df2)  # shouldn't raise an error since dtypes are equal

    # Dfs with different dtypes belonging to the same class
    df1 = df1.astype({"a": "float"})
    assert df1.dtypes["a"] != df2.dtypes["a"]
    assert_dtypes_equal(df1, df2)  # shouldn't raise an error since both are numeric

    # Dfs with different dtypes
    df2 = df2.astype("str")
    with pytest.raises(AssertionError):
        assert_dtypes_equal(sr1, sr2)

    # Dfs with categorical dtypes
    df1 = df1.astype("category")
    df2 = df2.astype("category")
    assert_dtypes_equal(df1, df2)  # shouldn't raise an error since both are categorical

    # Dfs with different dtypes (categorical and str)
    df1 = df1.astype({"a": "str"})
    with pytest.raises(AssertionError):
        assert_dtypes_equal(df1, df2)


def test_execute():
    data = np.random.rand(100, 64)
    modin_df, pandas_df = create_test_dfs(data)
    partitions = modin_df._query_compiler._modin_frame._partitions.flatten()
    mgr_cls = modin_df._query_compiler._modin_frame._partition_mgr_cls

    # check modin case
    with patch.object(mgr_cls, "wait_partitions", new=Mock()):
        modin.utils.execute(modin_df)
        mgr_cls.wait_partitions.assert_called_once()
        assert (mgr_cls.wait_partitions.call_args[0] == partitions).all()

    # check pandas case without error
    with patch.object(mgr_cls, "wait_partitions", new=Mock()):
        modin.utils.execute(pandas_df)
        mgr_cls.wait_partitions.assert_not_called()

    # muke sure `trigger_hdk_import=True` doesn't broke anything
    # when using other storage formats
    with patch.object(mgr_cls, "wait_partitions", new=Mock()):
        modin.utils.execute(modin_df, trigger_hdk_import=True)
        mgr_cls.wait_partitions.assert_called_once()

    # check several modin dataframes
    with patch.object(mgr_cls, "wait_partitions", new=Mock()):
        modin.utils.execute(modin_df, modin_df[modin_df.columns[:4]])
        mgr_cls.wait_partitions.assert_called
        assert mgr_cls.wait_partitions.call_count == 2



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Module houses class that wraps data (block partition) and its metadata."""

from typing import TYPE_CHECKING, Callable, Union

import pandas
import ray

if TYPE_CHECKING:
    from ray.util.client.common import ClientObjectRef

from modin.config import LazyExecution
from modin.core.dataframe.pandas.partitioning.partition import PandasDataframePartition
from modin.core.execution.ray.common import MaterializationHook, RayWrapper
from modin.core.execution.ray.common.deferred_execution import (
    DeferredExecution,
    MetaList,
    MetaListHook,
)
from modin.core.execution.ray.common.utils import ObjectIDType
from modin.logging import get_logger
from modin.pandas.indexing import compute_sliced_len
from modin.utils import _inherit_docstrings


class PandasOnRayDataframePartition(PandasDataframePartition):
    """
    The class implements the interface in ``PandasDataframePartition``.

    Parameters
    ----------
    data : ObjectIDType or DeferredExecution
        A reference to ``pandas.DataFrame`` that needs to be wrapped with this class
        or a reference to DeferredExecution that needs to be executed on demand.
    length : ObjectIDType or int, optional
        Length or reference to it of wrapped ``pandas.DataFrame``.
    width : ObjectIDType or int, optional
        Width or reference to it of wrapped ``pandas.DataFrame``.
    ip : ObjectIDType or str, optional
        Node IP address or reference to it that holds wrapped ``pandas.DataFrame``.
    meta : MetaList
        Meta information, containing the lengths and the worker address (the last value).
    meta_offset : int
        The lengths offset in the meta list.
    """

    execution_wrapper = RayWrapper

    def __init__(
        self,
        data: Union[ray.ObjectRef, "ClientObjectRef", DeferredExecution],
        length: int = None,
        width: int = None,
        ip: str = None,
        meta: MetaList = None,
        meta_offset: int = 0,
    ):
        super().__init__()
        if isinstance(data, DeferredExecution):
            data.subscribe()
        self._data_ref = data
        # The metadata is stored in the MetaList at 0 offset. If the data is
        # a DeferredExecution, the _meta will be replaced with the list, returned
        # by the remote function. The returned list may contain data for multiple
        # results and, in this case, _meta_offset corresponds to the meta related to
        # this partition.
        if meta is None:
            self._meta = MetaList([length, width, ip])
            self._meta_offset = 0
        else:
            self._meta = meta
            self._meta_offset = meta_offset

        log = get_logger()
        self._is_debug(log) and log.debug(
            "Partition ID: {}, Height: {}, Width: {}, Node IP: {}".format(
                self._identity,
                str(self._length_cache),
                str(self._width_cache),
                str(self._ip_cache),
            )
        )

    def __del__(self):
        """Unsubscribe from DeferredExecution."""
        if isinstance(self._data_ref, DeferredExecution):
            self._data_ref.unsubscribe()

    def apply(self, func: Union[Callable, ray.ObjectRef], *args, **kwargs):
        """
        Apply a function to the object wrapped by this partition.

        Parameters
        ----------
        func : callable or ray.ObjectRef
            A function to apply.
        *args : iterable
            Additional positional arguments to be passed in `func`.
        **kwargs : dict
            Additional keyword arguments to be passed in `func`.

        Returns
        -------
        PandasOnRayDataframePartition
            A new ``PandasOnRayDataframePartition`` object.

        Notes
        -----
        It does not matter if `func` is callable or an ``ray.ObjectRef``. Ray will
        handle it correctly either way. The keyword arguments are sent as a dictionary.
        """
        log = get_logger()
        self._is_debug(log) and log.debug(f"ENTER::Partition.apply::{self._identity}")
        de = DeferredExecution(self._data_ref, func, args, kwargs)
        data, meta, meta_offset = de.exec()
        self._is_debug(log) and log.debug(f"EXIT::Partition.apply::{self._identity}")
        return self.__constructor__(data, meta=meta, meta_offset=meta_offset)

    @_inherit_docstrings(PandasDataframePartition.add_to_apply_calls)
    def add_to_apply_calls(
        self,
        func: Union[Callable, ray.ObjectRef],
        *args,
        length=None,
        width=None,
        **kwargs,
    ):
        return self.__constructor__(
            data=DeferredExecution(self._data_ref, func, args, kwargs),
            length=length,
            width=width,
        )

    @_inherit_docstrings(PandasDataframePartition.drain_call_queue)
    def drain_call_queue(self):
        data = self._data_ref
        if not isinstance(data, DeferredExecution):
            return data

        log = get_logger()
        self._is_debug(log) and log.debug(
            f"ENTER::Partition.drain_call_queue::{self._identity}"
        )
        self._data_ref, self._meta, self._meta_offset = data.exec()
        self._is_debug(log) and log.debug(
            f"EXIT::Partition.drain_call_queue::{self._identity}"
        )

    @_inherit_docstrings(PandasDataframePartition.wait)
    def wait(self):
        self.drain_call_queue()
        RayWrapper.wait(self._data_ref)

    def __copy__(self):
        """
        Create a copy of this partition.

        Returns
        -------
        PandasOnRayDataframePartition
            A copy of this partition.
        """
        return self.__constructor__(
            self._data_ref,
            meta=self._meta,
            meta_offset=self._meta_offset,
        )

    def mask(self, row_labels, col_labels):
        """
        Lazily create a mask that extracts the indices provided.

        Parameters
        ----------
        row_labels : list-like, slice or label
            The row labels for the rows to extract.
        col_labels : list-like, slice or label
            The column labels for the columns to extract.

        Returns
        -------
        PandasOnRayDataframePartition
            A new ``PandasOnRayDataframePartition`` object.
        """
        log = get_logger()
        self._is_debug(log) and log.debug(f"ENTER::Partition.mask::{self._identity}")
        new_obj = super().mask(row_labels, col_labels)
        if isinstance(row_labels, slice) and isinstance(
            (len_cache := self._length_cache), ObjectIDType
        ):
            if row_labels == slice(None):
                # fast path - full axis take
                new_obj._length_cache = len_cache
            else:
                new_obj._length_cache = SlicerHook(len_cache, row_labels)
        if isinstance(col_labels, slice) and isinstance(
            (width_cache := self._width_cache), ObjectIDType
        ):
            if col_labels == slice(None):
                # fast path - full axis take
                new_obj._width_cache = width_cache
            else:
                new_obj._width_cache = SlicerHook(width_cache, col_labels)
        self._is_debug(log) and log.debug(f"EXIT::Partition.mask::{self._identity}")
        return new_obj

    @classmethod
    def put(cls, obj: pandas.DataFrame):
        """
        Put the data frame into Plasma store and wrap it with partition object.

        Parameters
        ----------
        obj : pandas.DataFrame
            A data frame to be put.

        Returns
        -------
        PandasOnRayDataframePartition
            A new ``PandasOnRayDataframePartition`` object.
        """
        return cls(cls.execution_wrapper.put(obj), len(obj.index), len(obj.columns))

    @classmethod
    def preprocess_func(cls, func):
        """
        Put a function into the Plasma store to use in ``apply``.

        Parameters
        ----------
        func : callable
            A function to preprocess.

        Returns
        -------
        ray.ObjectRef
            A reference to `func`.
        """
        return cls.execution_wrapper.put(func)

    def length(self, materialize=True):
        """
        Get the length of the object wrapped by this partition.

        Parameters
        ----------
        materialize : bool, default: True
            Whether to forcibly materialize the result into an integer. If ``False``
            was specified, may return a future of the result if it hasn't been
            materialized yet.

        Returns
        -------
        int or ray.ObjectRef
            The length of the object.
        """
        if (length := self._length_cache) is None:
            self.drain_call_queue()
            if (length := self._length_cache) is None:
                length, self._width_cache = _get_index_and_columns.remote(
                    self._data_ref
                )
                self._length_cache = length
        if materialize and isinstance(length, ObjectIDType):
            self._length_cache = length = RayWrapper.materialize(length)
        return length

    def width(self, materialize=True):
        """
        Get the width of the object wrapped by the partition.

        Parameters
        ----------
        materialize : bool, default: True
            Whether to forcibly materialize the result into an integer. If ``False``
            was specified, may return a future of the result if it hasn't been
            materialized yet.

        Returns
        -------
        int or ray.ObjectRef
            The width of the object.
        """
        if (width := self._width_cache) is None:
            self.drain_call_queue()
            if (width := self._width_cache) is None:
                self._length_cache, width = _get_index_and_columns.remote(
                    self._data_ref
                )
                self._width_cache = width
        if materialize and isinstance(width, ObjectIDType):
            self._width_cache = width = RayWrapper.materialize(width)
        return width

    def ip(self, materialize=True):
        """
        Get the node IP address of the object wrapped by this partition.

        Parameters
        ----------
        materialize : bool, default: True
            Whether to forcibly materialize the result into an integer. If ``False``
            was specified, may return a future of the result if it hasn't been
            materialized yet.

        Returns
        -------
        str
            IP address of the node that holds the data.
        """
        if (ip := self._ip_cache) is None:
            self.drain_call_queue()
        if materialize and isinstance(ip, ObjectIDType):
            self._ip_cache = ip = RayWrapper.materialize(ip)
        return ip

    @property
    def _data(self) -> Union[ray.ObjectRef, "ClientObjectRef"]:  # noqa: GL08
        self.drain_call_queue()
        return self._data_ref

    @property
    def _length_cache(self):  # noqa: GL08
        return self._meta[self._meta_offset]

    @_length_cache.setter
    def _length_cache(self, value):  # noqa: GL08
        self._meta[self._meta_offset] = value

    @property
    def _width_cache(self):  # noqa: GL08
        return self._meta[self._meta_offset + 1]

    @_width_cache.setter
    def _width_cache(self, value):  # noqa: GL08
        self._meta[self._meta_offset + 1] = value

    @property
    def _ip_cache(self):  # noqa: GL08
        return self._meta[-1]

    @_ip_cache.setter
    def _ip_cache(self, value):  # noqa: GL08
        self._meta[-1] = value


@ray.remote(num_returns=2)
def _get_index_and_columns(df):  # pragma: no cover
    """
    Get the number of rows and columns of a pandas DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        A pandas DataFrame which dimensions are needed.

    Returns
    -------
    int
        The number of rows.
    int
        The number of columns.
    """
    return len(df.index), len(df.columns)


PandasOnRayDataframePartition._eager_exec_func = PandasOnRayDataframePartition.apply
PandasOnRayDataframePartition._lazy_exec_func = (
    PandasOnRayDataframePartition.add_to_apply_calls
)


def _configure_lazy_exec(cls: LazyExecution):
    """Configure lazy execution mode for PandasOnRayDataframePartition."""
    mode = cls.get()
    get_logger().debug(f"Ray lazy execution mode: {mode}")
    if mode == "Auto":
        PandasOnRayDataframePartition.apply = (
            PandasOnRayDataframePartition._eager_exec_func
        )
        PandasOnRayDataframePartition.add_to_apply_calls = (
            PandasOnRayDataframePartition._lazy_exec_func
        )
    elif mode == "On":

        def lazy_exec(self, func, *args, **kwargs):
            return self._lazy_exec_func(func, *args, length=None, width=None, **kwargs)

        PandasOnRayDataframePartition.apply = lazy_exec
        PandasOnRayDataframePartition.add_to_apply_calls = (
            PandasOnRayDataframePartition._lazy_exec_func
        )
    elif mode == "Off":

        def eager_exec(self, func, *args, length=None, width=None, **kwargs):
            return self._eager_exec_func(func, *args, **kwargs)

        PandasOnRayDataframePartition.apply = (
            PandasOnRayDataframePartition._eager_exec_func
        )
        PandasOnRayDataframePartition.add_to_apply_calls = eager_exec
    else:
        raise ValueError(f"Invalid lazy execution mode: {mode}")


LazyExecution.subscribe(_configure_lazy_exec)


class SlicerHook(MaterializationHook):
    """
    Used by mask() for the slilced length computation.

    Parameters
    ----------
    ref : ObjectIDType
        Non-materialized length to be sliced.
    slc : slice
        The slice to be applied.
    """

    def __init__(self, ref: ObjectIDType, slc: slice):
        self.ref = ref
        self.slc = slc

    def pre_materialize(self):
        """
        Get the sliced length or object ref if not materialized.

        Returns
        -------
        int or ObjectIDType
        """
        if isinstance(self.ref, MetaListHook):
            len_or_ref = self.ref.pre_materialize()
            return (
                compute_sliced_len(self.slc, len_or_ref)
                if isinstance(len_or_ref, int)
                else len_or_ref
            )
        return self.ref

    def post_materialize(self, materialized):
        """
        Get the sliced length.

        Parameters
        ----------
        materialized : list or int

        Returns
        -------
        int
        """
        if isinstance(self.ref, MetaListHook):
            materialized = self.ref.post_materialize(materialized)
        return compute_sliced_len(self.slc, materialized)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import os
import unittest.mock
import warnings

import pytest
from packaging import version

import modin.config as cfg
from modin.config.envvars import _check_vars
from modin.config.pubsub import _UNSET, ExactStr


def reset_vars(*vars: tuple[cfg.Parameter]):
    """
    Reset value for the passed parameters.

    Parameters
    ----------
    *vars : tuple[Parameter]
    """
    for var in vars:
        var._value = _UNSET
        _ = os.environ.pop(var.varname, None)


@pytest.fixture
def make_unknown_env():
    varname = "MODIN_UNKNOWN"
    os.environ[varname] = "foo"
    yield varname
    del os.environ[varname]


@pytest.fixture(params=[str, ExactStr])
def make_custom_envvar(request):
    class CustomVar(cfg.EnvironmentVariable, type=request.param):
        """custom var"""

        default = 10
        varname = "MODIN_CUSTOM"
        choices = (1, 5, 10)

    return CustomVar


@pytest.fixture
def set_custom_envvar(make_custom_envvar):
    os.environ[make_custom_envvar.varname] = "  custom  "
    yield "Custom" if make_custom_envvar.type is str else "  custom  "
    del os.environ[make_custom_envvar.varname]


def test_unknown(make_unknown_env):
    with pytest.warns(UserWarning, match=f"Found unknown .*{make_unknown_env}.*"):
        _check_vars()


def test_custom_default(make_custom_envvar):
    assert make_custom_envvar.get() == 10


def test_custom_set(make_custom_envvar, set_custom_envvar):
    assert make_custom_envvar.get() == set_custom_envvar


def test_custom_help(make_custom_envvar):
    assert "MODIN_CUSTOM" in make_custom_envvar.get_help()
    assert "custom var" in make_custom_envvar.get_help()


def test_doc_module():
    import pandas

    import modin.pandas as pd
    from modin.config import DocModule

    DocModule.put("modin.config.test.docs_module")

    # Test for override
    assert (
        pd.DataFrame.apply.__doc__
        == "This is a test of the documentation module for DataFrame."
    )
    # Test for pandas doc when method is not defined on the plugin module
    assert pandas.DataFrame.isna.__doc__ in pd.DataFrame.isna.__doc__
    assert pandas.DataFrame.isnull.__doc__ in pd.DataFrame.isnull.__doc__
    # Test for override
    assert (
        pd.Series.isna.__doc__
        == "This is a test of the documentation module for Series."
    )
    # Test for pandas doc when method is not defined on the plugin module
    assert pandas.Series.isnull.__doc__ in pd.Series.isnull.__doc__
    assert pandas.Series.apply.__doc__ in pd.Series.apply.__doc__
    # Test for override
    assert pd.read_csv.__doc__ == "Test override for functions on the module."
    # Test for pandas doc when function is not defined on module.
    assert pandas.read_table.__doc__ in pd.read_table.__doc__


def test_hdk_envvar():
    try:
        import pyhdk

        defaults = cfg.HdkLaunchParameters.get()
        assert defaults["enable_union"] == 1
        if version.parse(pyhdk.__version__) >= version.parse("0.6.1"):
            assert defaults["log_dir"] == "pyhdk_log"
        del cfg.HdkLaunchParameters._value
    except ImportError:
        # This test is intended to check pyhdk internals. If pyhdk is not available, skip the version check test.
        pass

    os.environ[cfg.HdkLaunchParameters.varname] = "enable_union=2,enable_thrift_logs=3"
    params = cfg.HdkLaunchParameters.get()
    assert params["enable_union"] == 2
    assert params["enable_thrift_logs"] == 3

    os.environ[cfg.HdkLaunchParameters.varname] = "unsupported=X"
    del cfg.HdkLaunchParameters._value
    params = cfg.HdkLaunchParameters.get()
    assert params["unsupported"] == "X"
    try:
        import pyhdk

        pyhdk.buildConfig(**cfg.HdkLaunchParameters.get())
    except RuntimeError as e:
        assert str(e) == "unrecognised option '--unsupported'"
    except ImportError:
        # This test is intended to check pyhdk internals. If pyhdk is not available, skip the version check test.
        pass

    os.environ[cfg.HdkLaunchParameters.varname] = (
        "enable_union=4,enable_thrift_logs=5,enable_lazy_dict_materialization=6"
    )
    del cfg.HdkLaunchParameters._value
    params = cfg.HdkLaunchParameters.get()
    assert params["enable_union"] == 4
    assert params["enable_thrift_logs"] == 5
    assert params["enable_lazy_dict_materialization"] == 6


@pytest.mark.parametrize(
    "deprecated_var, new_var",
    [
        (cfg.ExperimentalGroupbyImpl, cfg.RangePartitioningGroupby),
        (cfg.ExperimentalNumPyAPI, cfg.ModinNumpy),
    ],
)
def test_deprecated_bool_vars_warnings(deprecated_var, new_var):
    """Test that deprecated parameters are raising `FutureWarnings` and their replacements don't."""
    old_depr_val = deprecated_var.get()
    old_new_var = new_var.get()

    try:
        reset_vars(deprecated_var, new_var)
        with pytest.warns(FutureWarning):
            deprecated_var.get()

        with pytest.warns(FutureWarning):
            deprecated_var.put(False)

        with unittest.mock.patch.dict(os.environ, {deprecated_var.varname: "1"}):
            with pytest.warns(FutureWarning):
                _check_vars()

        # check that the new var doesn't raise any warnings
        reset_vars(deprecated_var, new_var)
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            new_var.get()
            new_var.put(False)
            with unittest.mock.patch.dict(os.environ, {new_var.varname: "1"}):
                _check_vars()
    finally:
        deprecated_var.put(old_depr_val)
        new_var.put(old_new_var)


@pytest.mark.parametrize(
    "deprecated_var, new_var",
    [
        (cfg.ExperimentalGroupbyImpl, cfg.RangePartitioningGroupby),
        (cfg.ExperimentalNumPyAPI, cfg.ModinNumpy),
    ],
)
@pytest.mark.parametrize("get_depr_first", [True, False])
def test_deprecated_bool_vars_equals(deprecated_var, new_var, get_depr_first):
    """
    Test that deprecated parameters always have values equal to the new replacement parameters.

    Parameters
    ----------
    deprecated_var : EnvironmentVariable
    new_var : EnvironmentVariable
    get_depr_first : bool
        Defines an order in which the ``.get()`` method should be called when comparing values.
        If ``True``: get deprecated value at first ``deprecated_var.get() == new_var.get() == value``.
        If ``False``: get new value at first ``new_var.get() == deprecated_var.get() == value``.
        The logic of the ``.get()`` method depends on which parameter was initialized first,
        that's why it's worth testing both cases.
    """
    old_depr_val = deprecated_var.get()
    old_new_var = new_var.get()

    def get_values():
        return (
            (deprecated_var.get(), new_var.get())
            if get_depr_first
            else (new_var.get(), deprecated_var.get())
        )

    try:
        # case1: initializing the value using 'deprecated_var'
        reset_vars(deprecated_var, new_var)
        deprecated_var.put(True)
        val1, val2 = get_values()
        assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

        new_var.put(False)
        val1, val2 = get_values()
        assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

        new_var.put(True)
        val1, val2 = get_values()
        assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

        deprecated_var.put(False)
        val1, val2 = get_values()
        assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

        # case2: initializing the value using 'new_var'
        reset_vars(deprecated_var, new_var)
        new_var.put(True)
        val1, val2 = get_values()
        assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

        deprecated_var.put(False)
        val1, val2 = get_values()
        assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

        deprecated_var.put(True)
        val1, val2 = get_values()
        assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

        new_var.put(False)
        val1, val2 = get_values()
        assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

        # case3: initializing the value using 'deprecated_var' with env variable
        reset_vars(deprecated_var, new_var)
        with unittest.mock.patch.dict(os.environ, {deprecated_var.varname: "True"}):
            val1, val2 = get_values()
            assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

            new_var.put(False)
            val1, val2 = get_values()
            assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

            new_var.put(True)
            val1, val2 = get_values()
            assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

            deprecated_var.put(False)
            val1, val2 = get_values()
            assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

        # case4: initializing the value using 'new_var' with env variable
        reset_vars(deprecated_var, new_var)
        with unittest.mock.patch.dict(os.environ, {new_var.varname: "True"}):
            val1, val2 = get_values()
            assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

            deprecated_var.put(False)
            val1, val2 = get_values()
            assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)

            deprecated_var.put(True)
            val1, val2 = get_values()
            assert val1 == val2 == True  # noqa: E712 ('obj == True' comparison)

            new_var.put(False)
            val1, val2 = get_values()
            assert val1 == val2 == False  # noqa: E712 ('obj == False' comparison)
    finally:
        deprecated_var.put(old_depr_val)
        new_var.put(old_new_var)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import io
import warnings

import matplotlib
import numpy as np
import pandas
import pytest

import modin.pandas as pd
from modin.config import NPartitions
from modin.pandas.test.utils import (
    RAND_HIGH,
    RAND_LOW,
    create_test_dfs,
    df_equals,
    eval_general,
    random_state,
    test_data,
    test_data_keys,
    test_data_values,
)
from modin.pandas.utils import SET_DATAFRAME_ATTRIBUTE_WARNING
from modin.test.test_utils import warns_that_defaulting_to_pandas

NPartitions.put(4)

# Force matplotlib to not use any Xwindows backend.
matplotlib.use("Agg")


@pytest.mark.parametrize("method", ["items", "iterrows"])
def test_items_iterrows(method):
    data = test_data["float_nan_data"]
    modin_df, pandas_df = pd.DataFrame(data), pandas.DataFrame(data)

    for modin_item, pandas_item in zip(
        getattr(modin_df, method)(), getattr(pandas_df, method)()
    ):
        modin_index, modin_series = modin_item
        pandas_index, pandas_series = pandas_item
        df_equals(pandas_series, modin_series)
        assert pandas_index == modin_index


@pytest.mark.parametrize("name", [None, "NotPandas"])
def test_itertuples_name(name):
    data = test_data["float_nan_data"]
    modin_df, pandas_df = pd.DataFrame(data), pandas.DataFrame(data)

    modin_it_custom = modin_df.itertuples(name=name)
    pandas_it_custom = pandas_df.itertuples(name=name)
    for modin_row, pandas_row in zip(modin_it_custom, pandas_it_custom):
        np.testing.assert_equal(modin_row, pandas_row)


def test_itertuples_multiindex():
    data = test_data["int_data"]
    modin_df, pandas_df = pd.DataFrame(data), pandas.DataFrame(data)

    new_idx = pd.MultiIndex.from_tuples(
        [(i // 4, i // 2, i) for i in range(len(modin_df.columns))]
    )
    modin_df.columns = new_idx
    pandas_df.columns = new_idx
    modin_it_custom = modin_df.itertuples()
    pandas_it_custom = pandas_df.itertuples()
    for modin_row, pandas_row in zip(modin_it_custom, pandas_it_custom):
        np.testing.assert_equal(modin_row, pandas_row)


def test___iter__():
    modin_df = pd.DataFrame(test_data_values[0])
    pandas_df = pandas.DataFrame(test_data_values[0])

    modin_iterator = modin_df.__iter__()

    # Check that modin_iterator implements the iterator interface
    assert hasattr(modin_iterator, "__iter__")
    assert hasattr(modin_iterator, "next") or hasattr(modin_iterator, "__next__")

    pd_iterator = pandas_df.__iter__()
    assert list(modin_iterator) == list(pd_iterator)


@pytest.mark.parametrize("data", test_data_values, ids=test_data_keys)
def test___contains__(request, data):
    modin_df = pd.DataFrame(data)
    pandas_df = pandas.DataFrame(data)

    result = False
    key = "Not Exist"
    assert result == modin_df.__contains__(key)
    assert result == (key in modin_df)

    if "empty_data" not in request.node.name:
        result = True
        key = pandas_df.columns[0]
        assert result == modin_df.__contains__(key)
        assert result == (key in modin_df)


@pytest.mark.parametrize("expand_frame_repr", [False, True])
@pytest.mark.parametrize(
    "max_rows_columns",
    [(5, 5), (10, 10), (50, 50), (51, 51), (52, 52), (75, 75), (None, None)],
)
@pytest.mark.parametrize("frame_size", [101, 102])
def test_display_options_for___repr__(max_rows_columns, expand_frame_repr, frame_size):
    frame_data = random_state.randint(
        RAND_LOW, RAND_HIGH, size=(frame_size, frame_size)
    )
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)

    context_arg = [
        "display.max_rows",
        max_rows_columns[0],
        "display.max_columns",
        max_rows_columns[1],
        "display.expand_frame_repr",
        expand_frame_repr,
    ]
    with pd.option_context(*context_arg):
        modin_df_repr = repr(modin_df)
    with pandas.option_context(*context_arg):
        pandas_df_repr = repr(pandas_df)
    assert modin_df_repr == pandas_df_repr


def test___finalize__():
    data = test_data_values[0]
    with warns_that_defaulting_to_pandas():
        pd.DataFrame(data).__finalize__(None)


@pytest.mark.parametrize("data", test_data_values, ids=test_data_keys)
def test___copy__(data):
    modin_df = pd.DataFrame(data)
    pandas_df = pandas.DataFrame(data)

    modin_df_copy, pandas_df_copy = modin_df.__copy__(), pandas_df.__copy__()
    df_equals(modin_df_copy, pandas_df_copy)


@pytest.mark.parametrize("data", test_data_values, ids=test_data_keys)
def test___deepcopy__(data):
    modin_df = pd.DataFrame(data)
    pandas_df = pandas.DataFrame(data)

    modin_df_copy, pandas_df_copy = (
        modin_df.__deepcopy__(),
        pandas_df.__deepcopy__(),
    )
    df_equals(modin_df_copy, pandas_df_copy)


def test___repr__():
    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(1000, 100))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)
    assert repr(pandas_df) == repr(modin_df)

    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(1000, 99))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)
    assert repr(pandas_df) == repr(modin_df)

    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(1000, 101))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)
    assert repr(pandas_df) == repr(modin_df)

    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(1000, 102))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)
    assert repr(pandas_df) == repr(modin_df)

    # ___repr___ method has a different code path depending on
    # whether the number of rows is >60; and a different code path
    # depending on the number of columns is >20.
    # Previous test cases already check the case when cols>20
    # and rows>60. The cases that follow exercise the other three
    # combinations.
    # rows <= 60, cols > 20
    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(10, 100))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)

    assert repr(pandas_df) == repr(modin_df)

    # rows <= 60, cols <= 20
    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(10, 10))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)

    assert repr(pandas_df) == repr(modin_df)

    # rows > 60, cols <= 20
    frame_data = random_state.randint(RAND_LOW, RAND_HIGH, size=(100, 10))
    pandas_df = pandas.DataFrame(frame_data)
    modin_df = pd.DataFrame(frame_data)

    assert repr(pandas_df) == repr(modin_df)

    # Empty
    pandas_df = pandas.DataFrame(columns=["col{}".format(i) for i in range(100)])
    modin_df = pd.DataFrame(columns=["col{}".format(i) for i in range(100)])

    assert repr(pandas_df) == repr(modin_df)

    # From Issue #1705
    string_data = """"time","device_id","lat","lng","accuracy","activity_1","activity_1_conf","activity_2","activity_2_conf","activity_3","activity_3_conf"
"2016-08-26 09:00:00.206",2,60.186805,24.821049,33.6080017089844,"STILL",75,"IN_VEHICLE",5,"ON_BICYCLE",5
"2016-08-26 09:00:05.428",5,60.192928,24.767222,5,"WALKING",62,"ON_BICYCLE",29,"RUNNING",6
"2016-08-26 09:00:05.818",1,60.166382,24.700443,3,"WALKING",75,"IN_VEHICLE",5,"ON_BICYCLE",5
"2016-08-26 09:00:15.816",1,60.166254,24.700671,3,"WALKING",75,"IN_VEHICLE",5,"ON_BICYCLE",5
"2016-08-26 09:00:16.413",5,60.193055,24.767427,5,"WALKING",85,"ON_BICYCLE",15,"UNKNOWN",0
"2016-08-26 09:00:20.578",3,60.152996,24.745216,3.90000009536743,"STILL",69,"IN_VEHICLE",31,"UNKNOWN",0"""
    pandas_df = pandas.read_csv(io.StringIO(string_data))
    with warns_that_defaulting_to_pandas():
        modin_df = pd.read_csv(io.StringIO(string_data))
    assert repr(pandas_df) == repr(modin_df)


def test___repr__does_not_raise_attribute_column_warning():
    # See https://github.com/modin-project/modin/issues/5380
    df = pd.DataFrame([1])
    with warnings.catch_warnings():
        warnings.filterwarnings(action="error", message=SET_DATAFRAME_ATTRIBUTE_WARNING)
        repr(df)


@pytest.mark.parametrize("data", test_data_values, ids=test_data_keys)
def test_inplace_series_ops(data):
    pandas_df = pandas.DataFrame(data)
    modin_df = pd.DataFrame(data)

    if len(modin_df.columns) > len(pandas_df.columns):
        col0 = modin_df.columns[0]
        col1 = modin_df.columns[1]
        pandas_df[col1].dropna(inplace=True)
        modin_df[col1].dropna(inplace=True)
        df_equals(modin_df, pandas_df)

        pandas_df[col0].fillna(0, inplace=True)
        modin_df[col0].fillna(0, inplace=True)
        df_equals(modin_df, pandas_df)


# Note: Tests setting an attribute that is not an existing column label
def test___setattr__not_column():
    pandas_df = pandas.DataFrame([1, 2, 3])
    modin_df = pd.DataFrame([1, 2, 3])

    pandas_df.new_col = [4, 5, 6]
    modin_df.new_col = [4, 5, 6]

    df_equals(modin_df, pandas_df)

    # While `new_col` is not a column of the dataframe,
    # it should be accessible with __getattr__.
    assert modin_df.new_col == pandas_df.new_col


def test___setattr__mutating_column():
    # Use case from issue #4577
    pandas_df = pandas.DataFrame([[1]], columns=["col0"])
    modin_df = pd.DataFrame([[1]], columns=["col0"])

    # Replacing a column with a list should mutate the column in place.
    pandas_df.col0 = [3]
    modin_df.col0 = [3]

    df_equals(modin_df, pandas_df)
    # Check that the col0 attribute reflects the value update.
    df_equals(modin_df.col0, pandas_df.col0)

    pandas_df.col0 = pandas.Series([5])
    modin_df.col0 = pd.Series([5])

    # Check that the col0 attribute reflects this update
    df_equals(modin_df, pandas_df)

    pandas_df.loc[0, "col0"] = 4
    modin_df.loc[0, "col0"] = 4

    # Check that the col0 attribute reflects update via loc
    df_equals(modin_df, pandas_df)
    assert modin_df.col0.equals(modin_df["col0"])

    # Check that attempting to add a new col via attributes raises warning
    # and adds the provided list as a new attribute and not a column.
    with pytest.warns(
        UserWarning,
        match=SET_DATAFRAME_ATTRIBUTE_WARNING,
    ):
        modin_df.col1 = [4]

    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="error",
            message=SET_DATAFRAME_ATTRIBUTE_WARNING,
        )
        modin_df.col1 = [5]
        modin_df.new_attr = 6
        modin_df.col0 = 7

    assert "new_attr" in dir(
        modin_df
    ), "Modin attribute was not correctly added to the df."
    assert (
        "new_attr" not in modin_df
    ), "New attribute was not correctly added to columns."
    assert modin_df.new_attr == 6, "Modin attribute value was set incorrectly."
    assert isinstance(
        modin_df.col0, pd.Series
    ), "Scalar was not broadcasted properly to an existing column."


@pytest.mark.parametrize("data", test_data_values, ids=test_data_keys)
def test_isin(data):
    pandas_df = pandas.DataFrame(data)
    modin_df = pd.DataFrame(data)

    val = [1, 2, 3, 4]
    pandas_result = pandas_df.isin(val)
    modin_result = modin_df.isin(val)

    df_equals(modin_result, pandas_result)


def test_isin_with_modin_objects():
    modin_df1, pandas_df1 = create_test_dfs({"a": [1, 2], "b": [3, 4]})
    modin_series, pandas_series = pd.Series([1, 4, 5, 6]), pandas.Series([1, 4, 5, 6])

    eval_general(
        (modin_df1, modin_series),
        (pandas_df1, pandas_series),
        lambda srs: srs[0].isin(srs[1]),
    )

    modin_df2 = modin_series.to_frame("a")
    pandas_df2 = pandas_series.to_frame("a")

    eval_general(
        (modin_df1, modin_df2),
        (pandas_df1, pandas_df2),
        lambda srs: srs[0].isin(srs[1]),
    )

    # Check case when indices are not matching
    modin_df1, pandas_df1 = create_test_dfs({"a": [1, 2], "b": [3, 4]}, index=[10, 11])

    eval_general(
        (modin_df1, modin_series),
        (pandas_df1, pandas_series),
        lambda srs: srs[0].isin(srs[1]),
    )
    eval_general(
        (modin_df1, modin_df2),
        (pandas_df1, pandas_df2),
        lambda srs: srs[0].isin(srs[1]),
    )
